"""
    domonic.geom.vec4
    ====================================
    written by.ai
"""
import math
from domonic.javascript import Math


class vec4():
    """[vec4]
    """

    def __init__(self, x: float = 0.0, y: float = 0.0, z: float = 0.0, w: float = 0.0, n: float = 1.0):
        self.x: float = x
        self.y: float = y
        self.z: float = z
        self.w: float = w
        self.n: float = n

    def __add__(self, other):
        x = self.x + other.x
        y = self.y + other.y
        z = self.z + other.z
        w = self.w + other.w
        return vec4(x, y, z, w)

    def __sub__(self, other):
        x = self.x - other.x
        y = self.y - other.y
        z = self.z - other.z
        w = self.w - other.w
        return vec4(x, y, z, w)

    def __mul__(self, other):
        n = self.n * other.n
        a = self.x * other.x + self.y * other.y + self.z * other.z
        b = self.x * other.y - self.y * other.x + self.z * other.w
        c = self.x * other.z - self.z * other.x + self.y * other.w
        d = self.w * other.n - self.x * other.w
        x = a + b * n
        y = b - a * n
        z = c + d * n
        w = d - c * n
        return vec4(x, y, z, w)



"""
    domonic CLI
    ====================================
    - some useful cli commands
"""

import argparse
import os
import sys

prog = '''

function project(){
    PROJECT_NAME=$1
    mkdir $PROJECT_NAME
    cd $PROJECT_NAME

    mkdir static
    mkdir static/js
    mkdir static/css
    mkdir static/img
    mkdir static/data

    mkdir archive
    touch app.py
    echo "from domonic.html import *" >> app.py

    touch Makefile

    touch README.md
    echo "# $PROJECT_NAME" >> README.md
    echo "## Description" >> README.md
    echo "## Installation" >> README.md
    echo "## Usage" >> README.md
    echo "## Tests" >> README.md
    echo "## License" >> README.md

    mkdir app
    touch app/__init__.py
    echo '__version__ = "0.0.1"' >> app/__init__.py

    git init
    touch .gitignore
    echo "*.pyc" >> .gitignore
    echo "*.pyo" >> .gitignore
    echo "*.swp" >> .gitignore
    echo "*.swo" >> .gitignore
    echo "*.DS_Store" >> .gitignore
    echo "__pycache__/" >> .gitignore

    touch static/js/master.js
    touch static/css/styles.css
    touch static/data/data.json

    python3 -m venv venv
    . venv/bin/activate

    pip3 install requests
    pip3 install sanic
    pip3 install domonic
    pip3 freeze >> requirements.txt

    chmod -R 777 static
    open .
}

'''
# TODO - nautilus instead of open for linux
# wrewrite as pure python


def project(name):
    """
    this will replace the older bash script
    """
    PROJECT_NAME = name
    os.mkdir(PROJECT_NAME)
    os.chdir(PROJECT_NAME)

    # create a Makefile
    # os.system("touch Makefile")
    with open("Makefile", "w") as f:
        # start venv and run app
        # f.write(". venv/bin/activate\n")
        # f.write("python3 app.py\n")
        # TOD as a run command
        f.write("""
run:
\t(. venv/bin/activate; python3 app.py;)
""")

    # create a README.md
    with open("README.md", "w") as f:
        f.write("# " + PROJECT_NAME + "\n")
        f.write("## Description" + "\n")
        f.write("## Installation" + "\n")
        f.write("## Usage" + "\n")
        f.write("## Tests" + "\n")
        f.write("## License")

    # create app
    os.mkdir("app")
    with open("app/__init__.py", "w") as f:
        f.write('__version__ = "0.0.1"')

    # create a git repo
    os.system("git init")
    with open(".gitignore", "w") as f:
        f.write("*.pyc")
        f.write("*.pyo")
        f.write("*.swp")
        f.write("*.swo")
        f.write("*.DS_Store")
        f.write("__pycache__/")

    # create a venv
    # os.system("python3 -m venv venv")
    # if os.name == "nt":
    #     os.system("venv\Scripts\activate")
    # else:
    #     os.system("source venv/bin/activate")
    # # install requirements
    # os.system("python3 -m pip install requests")
    # os.system("python3 -m pip install sanic")
    # os.system("python3 -m pip install domonic")
    # os.system("python3 -m pip freeze > requirements.txt")

    # ask the user which server they want to use
    server_opt = ["none", "sanic", "flask", "cherrypy", "django", "bottle", "pyramid", "werkzeug", "tornado", "aiohttp", "fastapi", "starlette"]
    print("You want a server?")
    for i, server in enumerate(server_opt):
        print(str(i) + ": " + server)
    server_choice = input("Enter a number: ")
    try:
        server_choice = server_opt[int(server_choice)]
    except ValueError:
        if server_choice in server_opt:
            server_choice = server_choice

    # TODO - any plugins?.. cors etc

    # with python not touch
    with open("app.py", "w") as f:
        # write the hello world for the given server
        from domonic.ext import get_hello_world
        code = get_hello_world(server_choice)
        if code is not None:
            f.write(code)
        else:
            f.write("from domonic.html import *")

    os.system("python3 -m venv venv")
    from domonic.utils import Utils
    if Utils.is_windows():
        # create a bat file to activate the venv an install requirements
        with open("activate.bat", "w") as f:
            f.write("@echo off\n")
            f.write("call \"venv\Scripts\activate\"\n")
            f.write("python3 -m pip install requests\n")
            if server_choice != 'none':
                f.write(f"python3 -m pip install {server_choice}\n")
            f.write("python3 -m pip install domonic\n")
            f.write("python3 -m pip freeze > requirements.txt\n")
        # call the bat file
        os.system("activate.bat")
        # remove the activate.bat file
        os.system("del activate.bat")
    else:
        # create a bash file to activate the venv an install requirements
        with open("activate.sh", "w") as f:
            f.write("#!/bin/bash\n")
            f.write("source venv/bin/activate\n")
            f.write("python3 -m pip install requests\n")
            if server_choice != 'none':
                f.write(f"python3 -m pip install {server_choice}\n")
            f.write("python3 -m pip install domonic\n")
            f.write("python3 -m pip freeze > requirements.txt\n")
        # call the bash file
        os.system("bash activate.sh")
        # remove the activate.sh file
        os.system("rm activate.sh")

    # license_opt = ["none", "mit", "gpl", "apache", "bsd", "mpl"]
    # print("You want a license?")
    # for i, license in enumerate(license_opt):
    #     print(str(i) + ": " + license)
    # license_choice = input("Enter a number: ")
    # license_choice = license_opt[int(license_choice)]
    # dl the license

    # create static
    # TODO - build asset folders based on server choice.
    os.mkdir("static")
    os.mkdir("static/js")
    os.mkdir("static/css")
    os.mkdir("static/img")
    os.mkdir("static/data")

    # create files
    with open("static/js/master.js", "w") as f:
        f.write("")
    with open("static/css/styles.css", "w") as f:
        f.write("")
    with open("static/data/data.json", "w") as f:
        f.write("")

    # chmod
    if os.name == 'posix':
        os.system("chmod -R 777 static")

    if Utils.is_mac():
        os.system("open .")
    elif Utils.is_linux():
        os.system("nautilus .")
    elif Utils.is_windows():
        # TODO - check what to do on windows
        os.system("start .")
        # explorer.exe .?
        # os.system("explorer.exe .")

# def webpage(content):
#     from domonic.components import webpage_tmpl
#     with open("index.html", "w") as f:
#         f.write(webpage_tmpl(content))


def parse_args():
    parser = argparse.ArgumentParser(add_help=False,
                                     prog="domonic",
                                     usage="%(prog)s [options]",
                                     description="Generate HTML with Python 3")
    parser.add_argument('-h', '--help', help="Opens the online docs in your default browser", action='store_true')
    parser.add_argument('-v', '--version', action='store_true')
    parser.add_argument('-p', '--project', help="Create a new project", type=str)
    parser.add_argument('-e', '--eval', help="Evaluates a domonic pyml string and returns html", type=str)  # default=sys.stdin, nargs='*')

    parser.add_argument('-a', '--assets', help="Generate an assets directory with common files", action='store_true')
    parser.add_argument('-d', '--download', help="Attempts to to generate domonic template from a webpage", type=str)

    # parser.add_argument('-u', '--ui', help="launches a UI")
    # parser.add_argument('-p', '--pyml2html', help="converts a .pyml template file to html", type=str)
    # parser.add_argument('-g', '--html2pyml', help="converts a .html file to a .pyml template", type=str)

    # parser.add_argument('-w', '--website', action='store_true')  # launch the docs
    # parser.add_argument('-s', '--server', help="runs python -m http.server", type=str)
    # parser.add_argument('-u', '--url', help="url to launch the server", type=str)

    # parser.add_argument('-j', '--csv2json', help="converts a csv file to a json file", type=str)
    # parser.add_argument('-c', '--json2csv', help="converts a json file to a csv file", type=str)
    # parser.add_argument('-m', '--merge', help="merges two csv files", type=str)
    # parser.add_argument('-d', '--diff', help="compares two csv files", type=str)
    # parser.add_argument('-t', '--table', help="creates a table from a csv file", type=str)
    # parser.add_argument('-r', '--replace', help="replaces a value in a csv file", type=str)

    # parser.add_argument('-c', '--json2ini', help="converts a json file to an ini file", type=str)
    # parser.add_argument('-i', '--ini2json', help="converts an ini file to a json file", type=str)
    # parser.add_argument('-m', '--merge', help="merges two ini files", type=str)

    # -- ideas
    # -- change all file extensions. from, to
    # -- generate assets/app/license/readme/sitemap.

    args = parser.parse_args()
    return args


def do_things(arguments):
    from domonic.terminal import TerminalException
    try:
        if arguments.assets is True:
            from domonic.utils import Utils
            Utils.init_assets()
            # --license,readme,sitemap,requirements
    except TerminalException as e:
        print(e)

    # print(arguments.download)
    if arguments.download is not None:
        print('creating domonic template from url:')
        from domonic import domonic
        page = domonic.get(arguments.download)

        from domonic.html import render
        from domonic.utils import Utils
        print("filename:", Utils.url2file(arguments.download))
        render(page, Utils.url2file(arguments.download))

    if arguments.project is not None:
        print('creating a basic project:')
        project(arguments.project)

    if arguments.help is True:
        import webbrowser
        webbrowser.open_new("https://domonic.readthedocs.io/")

    if arguments.version is True:
        from domonic import __version__
        print(__version__)
        return __version__

    if arguments.eval is not None:
        import domonic
        result = f"{domonic.domonic.domonify(arguments.eval)}"
        print(result)
        return result

    # if arguments.server is True:
        # port = domonic.get(arguments.server)
        # os.system('python -m http.server ' + port)


def run():
    """[Entry point required by setup.py console_scripts. Saves having to add alias to .bash_profile]
    """
    args = parse_args()
    do_things(args)


if __name__ == "__main__":
    run()



"""
    domonic.events
    ====================================

    https://developer.mozilla.org/en-US/docs/Web/Events

"""

# from typing import *
import time


class EventTarget(object):
    """EventTarget is a class you can extend to give your obj event dispatching abilities"""

    def __init__(self, *args, **kwargs) -> None:
        self.listeners = {}

    def hasEventListener(self, _type: str) -> bool:
        return _type in self.listeners

    # TODO - event: str, function, useCapture: bool
    # def addEventListener(self, event: str, function, useCapture: bool) -> None:
    def addEventListener(self, _type: str, callback, *args, **kwargs):
        if _type not in self.listeners:
            self.listeners[_type] = []
        self.listeners[_type].append(callback)

    def removeEventListener(self, _type: str, callback):
        if _type not in self.listeners:
            return

        stack = self.listeners[_type]

        for thing in stack:
            if thing == callback:
                stack.remove(thing)
                return

    def dispatchEvent(self, event):
        if event.type not in self.listeners:
            return True  # huh?. surely false?

        stack = self.listeners[event.type]
        # .slice()
        event.target = (
            self  # TODO/NOTE - is this correct? - cant think where else would set it
        )

        for thing in stack:
            try:
                thing(event)
                # type(thing, (Event,), self)
            except Exception as e:
                print(e)
                thing()  # try calling without params, user may not create param

        return not event.defaultPrevented

    # async def dispatchEventAsync(self, event):
    #     if event.type not in self.listeners:
    #         return True

    #     stack = self.listeners[event.type]
    #     event.target = self
    #     for thing in stack:
    #         await thing(event)


EventDispatcher = EventTarget  #: legacy alias

# https://developer.mozilla.org/en-US/docs/Web/API/Server-sent_events
# class EventSource(EventTarget):
#     """ Baseclass for Node """
#     def __init__(self, *args, **kwargs):
#         self._readyState = "2"
#         self._url = ""
#         self._withCredentials = False
#         super(EventSource, self).__init__(*args, **kwargs)
#     @property
#     def readyState(self):
#         """ A number representing the state of the connection.
#         Possible values are CONNECTING (0), OPEN (1), or CLOSED (2). """
#         return self._readyState
#     @property
#     def url(self):
#         """ A DOMString representing the URL of the source. """
#         return self._url
#     @property
#     def withCredentials(self):
#         """ A boolean value indicating whether the EventSource object was
#           instantiated with cross-origin (CORS) credentials
#         set (true), or not (false, the default). """
#         return self._withCredentials
#     def close(self):
#         """ Closes the connection to the EventSource. """
#         self._readyState = "0"
#     def onreadystatechange(self, event):
#         """ Called when the state of the connection changes. """
#         if event.target.readyState == "0":
#             self._readyState = "0"
#         elif event.target.readyState == "1":
#             self._readyState = "1"
#         elif event.target.readyState == "2":
#             self._readyState = "2"
#         else:
#             self._readyState = "2"
#     def onmessage(self, event):
#         """ Called when a message is received. """
#         pass
#     def onerror(self, event):
#         """ Called when an error occurs. """
#         pass
#     def onopen(self, event):
#         """ Called when the connection is established. """
#         pass


class Event(object):
    """event"""

    EMPTIED: str = "emptied"  #:
    ABORT: str = "abort"  #:
    AFTERPRINT: str = "afterprint"  #:
    BEFOREPRINT: str = "beforeprint"  #:
    BEFOREUNLOAD: str = "beforeunload"  #:
    CANPLAY: str = "canplay"  #:
    CANPLAYTHROUGH: str = "canplaythrough"  #:
    CHANGE: str = "change"  #:
    DURATIONCHANGE: str = "durationchange"  #:
    ENDED: str = "ended"  #:
    ERROR: str = "error"  #:
    FULLSCREENCHANGE: str = "fullscreenchange"  #:
    FULLSCREENERROR: str = "fullscreenerror"  #:
    INPUT: str = "input"  #:
    INVALID: str = "invalid"  #:
    LOAD: str = "load"  #:
    LOADEDDATA: str = "loadeddata"  #:
    LOADEDMETADATA: str = "loadedmetadata"  #:
    MESSAGE: str = "message"  #:
    OFFLINE: str = "offline"  #:
    ONLINE: str = "online"  #:
    OPEN: str = "open"  #:
    PAUSE: str = "pause"  #:
    PLAY: str = "play"  #:
    PLAYING: str = "playing"  #:
    PROGRESS: str = "progress"  #:
    RATECHANGE: str = "ratechange"  #:
    READYSTATECHANGE: str = "readystatechange"  #:
    RESIZE: str = "resize"  #:
    RESET: str = "reset"  #:
    SCROLL: str = "scroll"  #:
    SEARCH: str = "search"  #:
    SEEKED: str = "seeked"  #:
    SEEKING: str = "seeking"  #:
    SELECT: str = "select"  #:
    SHOW: str = "show"  #:
    STALLED: str = "stalled"  #:
    SUBMIT: str = "submit"  #:
    SUSPEND: str = "suspend"  #:
    TOGGLE: str = "toggle"  #:
    UNLOAD: str = "unload"  #:
    VOLUMECHANGE: str = "volumechange"  #:
    WAITING: str = "waiting"  #:

    CAPTURING_PHASE: int = 1
    AT_TARGET: int = 2
    BUBBLING_PHASE: int = 3

    def __str__(self) -> str:
        return self.type + ":" + str(self.timeStamp)

    def __init__(self, _type: str, options: dict = None, *args, **kwargs) -> None:
        options = options or kwargs  # if options is none use kwargs
        self.type: str = _type
        self.bubbles: bool = options.get("bubbles", True)
        self.cancelable: bool = options.get("cancelable", True)
        self.cancelBubble: bool = options.get("cancelBubble", False)
        self.composed: bool = options.get("composed", True)
        self.currentTarget: object = options.get("currentTarget", None)
        self.defaultPrevented: bool = options.get("defaultPrevented", False)
        self.eventPhase: int = options.get("eventPhase", None)
        self.explicitOriginalTarget: object = options.get(
            "explicitOriginalTarget", None
        )
        self.isTrusted: bool = options.get("isTrusted", False)
        self.originalTarget: object = options.get("originalTarget", None)
        self.returnValue: bool = options.get("returnValue", True)
        self.srcElement: object = options.get("srcElement", None)
        self.target: object = options.get("target", None)
        # ms = time.time_ns() # 1000000 py3.7 up
        self.timeStamp: float = int(round(time.time() * 1000))

    def composedPath(self):
        return self.type + ":" + str(self.timeStamp)

    def initEvent(self, _type: str = None, *args, **kwargs) -> "Event":
        self.__init__(_type, args, kwargs)

    def stopPropagation(self):
        """[prevents further propagation of the current event in the capturing and bubbling phases]"""
        # self.defaultPrevented = True
        # self.returnValue = None
        # self.originalTarget = None
        # self.explicitOriginalTarget = None
        # self.target = None
        # self.srcElement = None
        # self.bubbles = None
        # self.cancelable = None
        # self.cancelBubble = None
        # self.composed = None
        # self.currentTarget = None
        # self.eventPhase = None
        # self.isTrusted = None
        # self.returnValue = None
        # self.timeStamp = int(round(time.time() * 1000))
        # self.type = None
        pass

    def msConvertURL(self):
        pass

    def preventDefault(self):
        pass

    def stopImmediatePropagation(self):
        pass


class UIEvent(Event):
    """UIEvent"""

    def __init__(self, _type: str, options: dict = None, *args, **kwargs) -> None:
        options = options or kwargs  # if options is none use kwargs
        self.detail = options.get("detail", None)
        self.view = options.get("view", None)
        self.detail = options.get("detail", None)
        self.layerX = options.get("layerX", None)
        self.layerY = options.get("layerY", None)
        # self.which = options.get("which", None)
        self.sourceCapabilities = options.get("sourceCapabilities", None)
        super().__init__(_type, options, *args, **kwargs)

    def initUIEvent(
        self, _type: str, canBubble: bool, cancelable: bool, view, detail
    ) -> "UIEvent":
        self._type = _type
        self.canBubble = canBubble
        self.cancelable = cancelable
        self.view = view
        self.detail = detail


class MouseEvent(UIEvent):
    """mouse events"""

    CLICK: str = "click"  #:
    CONTEXTMENU: str = "contextmenu"  #:
    DBLCLICK: str = "dblclick"  #:
    MOUSEDOWN: str = "mousedown"  #:
    MOUSEENTER: str = "mouseenter"  #:
    MOUSELEAVE: str = "mouseleave"  #:
    MOUSEMOVE: str = "mousemove"  #:
    MOUSEOVER: str = "mouseover"  #:
    MOUSEOUT: str = "mouseout"  #:
    MOUSEUP: str = "mouseup"  #:

    def __init__(self, _type: str, options: dict = None, *args, **kwargs) -> None:
        options = options or kwargs
        self.x = options.get("x", 0)
        self.y = options.get("y", 0)
        self._clientX = options.get("clientX", 0)
        self._clientX = options.get("clientY", 0)
        self._altKey: bool = options.get("altKey", False)
        self._ctrlKey: bool = options.get("ctrlKey", False)
        self._shiftKey: bool = options.get("shiftKey", False)
        self._metaKey: bool = options.get("metaKey", False)
        self._button = None
        self._buttons = []
        super().__init__(_type, options, *args, **kwargs)

    def initMouseEvent(
        self,
        _type: str = None,
        canBubble: bool = True,
        cancelable: bool = True,
        view=None,
        detail=None,
        screenX: float = 0,
        screenY: float = 0,
        clientX: float = 0,
        clientY: float = 0,
        ctrlKey: bool = False,
        altKey: bool = False,
        shiftKey: bool = False,
        metaKey: bool = False,
        button=None,
        relatedTarget=None,
        from_json={},
        *args,
        **kwargs
    ) -> "MouseEvent":
        # print('initMouseEvent')
        self._type = _type
        self.canBubble = canBubble
        self.cancelable = cancelable
        self.view = view
        self.detail = detail
        self.screenX = screenX
        self.screenY = screenY
        self._clientX = clientX
        self._clientY = clientY
        self._ctrlKey = ctrlKey
        self._altKey = altKey
        self._shiftKey = shiftKey
        self._metaKey = metaKey
        self._button = button
        self.relatedTarget = relatedTarget
        # TODO - parse from_json - so can relay

    @property
    def clientX(self):
        return self.x

    @property
    def clientY(self):
        return self.y

    @property
    def altKey(self):
        return self._altKey

    @property
    def ctrlKey(self):
        return self._ctrlKey

    @property
    def shiftKey(self):
        return self._shiftKey

    @property
    def metaKey(self):
        return self._metaKey

    @property
    def button(self):
        return self._button

    @property
    def buttons(self):
        return self._buttons

    @property
    def which(self):
        return self._button

    # MOUSE_EVENT
    def getModifierState(self):
        """Returns an array containing target ranges that will be affected by the insertion/deletion"""
        pass

    # MovementX Returns the horizontal coordinate of the mouse pointer relative to the position of the last mousemove event MouseEvent
    # MovementY Returns the vertical coordinate of the mouse pointer relative to the position of the last mousemove event   MouseEvent
    # offsetX   Returns the horizontal coordinate of the mouse pointer relative to the position of the edge of the target element   MouseEvent
    # offsetY   Returns the vertical coordinate of the mouse pointer relative to the position of the edge of the target element MouseEvent
    # pageX Returns the horizontal coordinate of the mouse pointer, relative to the document, when the mouse event was triggered    MouseEvent
    # pageY Returns the vertical coordinate of the mouse pointer, relative to the document, when the mouse event was triggered  MouseEvent
    # region        MouseEvent
    # relatedTarget Returns the element related to the element that triggered the mouse event   MouseEvent, FocusEvent


class KeyboardEvent(UIEvent):
    """keyboard events"""

    KEYDOWN: str = "keydown"  #:
    KEYPRESS: str = "keypress"  #:
    KEYUP: str = "keyup"  #:

    DOM_KEY_LOCATION_LEFT: int = 0  #:
    DOM_KEY_LOCATION_STANDARD: int = 1  #:
    DOM_KEY_LOCATION_RIGHT: int = 2  #:
    DOM_KEY_LOCATION_NUMPAD: int = 3  #:
    DOM_KEY_LOCATION_MOBILE: int = 4  #:
    DOM_KEY_LOCATION_JOYSTICK: int = 5  #:

    def __init__(self, _type: str, options: dict = None, *args, **kwargs) -> None:
        options = options or kwargs  # if options is none use kwargs
        self._altKey: bool = options.get("altKey", False)
        self._ctrlKey: bool = options.get("ctrlKey", False)
        self._shiftKey: bool = options.get("shiftKey", False)
        self._metaKey: bool = options.get("metaKey", False)

        self.charCode = options.get("charCode", None)
        self.code = options.get("code", None)
        self.key = options.get("key", None)
        self.keyCode = options.get("keyCode", None)

        super().__init__(_type, options, *args, **kwargs)

    def initKeyboardEvent(
        self,
        typeArg: str,
        canBubbleArg: bool,
        cancelableArg: bool,
        viewArg,
        charArg,
        keyArg,
        locationArg,
        modifiersListArg,
        repeat,
    ) -> "KeyboardEvent":
        self._type = typeArg
        self.canBubbleArg = canBubbleArg
        self.cancelableArg = cancelableArg
        self.viewArg = viewArg
        self.charArg = charArg
        self.keyArg = keyArg
        self.locationArg = locationArg
        self.modifiersListArg = modifiersListArg
        self.repeat = repeat

    @property
    def altKey(self):
        return self._altKey

    @property
    def ctrlKey(self):
        return self._ctrlKey

    @property
    def shiftKey(self):
        return self._shiftKey

    @property
    def metaKey(self):
        return self._metaKey

    @property
    def unicode(self):
        return self.key

    # @property
    # def keyCode(self):
    #     return self.keyCode

    # @property
    # def charCode(self):
    #     return self.charCode

    # @property
    # def code(self):
    #     return self.code

    # @property
    # def key(self):
    #     return self.key

    # def isComposing(self, *args, **kwargs):
    #     pass

    # KeyboardEvent
    # isComposing   Returns whether the state of the event is composing or not  InputEvent, KeyboardEvent
    # repeat    Returns whether a key is being hold down repeatedly, or not KeyboardEvent
    # location  Returns the location of a key on the keyboard or device KeyboardEvent


class CompositionEvent(UIEvent):
    """CompositionEvent"""

    START: str = "compositionstart"
    END: str = "compositionend"
    UPDATE: str = "compositionupdate"

    def __init__(self, _type: str, options: dict = None, *args, **kwargs) -> None:
        options = options or kwargs  # if options is none use kwargs
        self.data = options.get("data", None)
        self.locale = options.get("locale", None)
        super().__init__(_type, options, *args, **kwargs)


class FocusEvent(UIEvent):
    """FocusEvent"""

    BLUR: str = "blur"  #:
    FOCUS: str = "focus"  #:
    FOCUSIN: str = "focusin"  #:
    FOCUSOUT: str = "focusout"  #:

    def __init__(self, _type: str, options: dict = None, *args, **kwargs) -> None:
        options = options or kwargs  # if options is none use kwargs
        self.relatedTarget = options.get("relatedTarget", None)
        super().__init__(_type, options, *args, **kwargs)


class TouchEvent(UIEvent):
    """TouchEvent"""

    TOUCHCANCEL: str = "touchcancel"  #:
    TOUCHEND: str = "touchend"  #:
    TOUCHMOVE: str = "touchmove"  #:
    TOUCHSTART: str = "touchstart"  #:

    def __init__(self, _type: str, options: dict = None, *args, **kwargs) -> None:
        options = options or kwargs  # if options is none use kwargs
        self.shiftKey = options.get("shiftKey", False)
        self.altKey = options.get("altKey", False)
        self.changedTouches = options.get("changedTouches", [])
        self.ctrlKey = options.get("ctrlKey", False)
        self.metaKey = options.get("metaKey", False)
        self.shiftKey = options.get("shiftKey", False)
        self.targetTouches = options.get("targetTouches", [])
        self.touches = options.get("touches", [])
        super().__init__(_type, options, *args, **kwargs)


class WheelEvent(UIEvent):
    """WheelEvent"""

    MOUSEWHEEL: str = "mousewheel"  # DEPRECATED - USE WHEEL  #:
    WHEEL: str = "wheel"  #:

    def __init__(self, _type: str, options: dict = None, *args, **kwargs) -> None:
        options = options or kwargs  # if options is none use kwargs
        self.deltaX = options.get("deltaX", 0)
        self.deltaY = options.get("deltaY", 0)
        self.deltaZ = options.get("deltaZ", 0)
        self.deltaMode = options.get("deltaMode", 0)
        super().__init__(_type, options, *args, **kwargs)


class AnimationEvent(Event):
    """AnimationEvent"""

    ANIMATIONEND: str = "animationend"  #:
    ANIMATIONITERATION: str = "animationiteration"  #:
    ANIMATIONSTART: str = "animationstart"  #:

    def __init__(self, _type: str, options: dict = None, *args, **kwargs) -> None:
        options = options or kwargs  # if options is none use kwargs
        self.animationName = options.get("animationName", None)
        """ Returns the name of the animation """
        self.elapsedTime = options.get("elapsedTime", None)
        """ Returns the number of seconds an animation has been running """
        self.pseudoElement = options.get("pseudoElement", None)
        """ Returns the name of the pseudo-element of the animation """
        super().__init__(_type, options, *args, **kwargs)


class ClipboardEvent(Event):
    """ClipboardEvent"""

    COPY: str = "copy"  #:
    CUT: str = "cut"  #:
    PASTE: str = "paste"  #:

    def __init__(self, _type: str, options: dict = None, *args, **kwargs) -> None:
        options = options or kwargs  # if options is none use kwargs
        self.clipboardData = options.get("clipboardData", None)
        """ Returns an object containing the data affected by the clipboard operation """
        super().__init__(_type, options, *args, **kwargs)


class ErrorEvent(Event):
    """ErrorEvent"""

    ERROR: str = "error"  #:

    def __init__(self, _type: str, options: dict = None, *args, **kwargs) -> None:
        options = options or kwargs  # if options is none use kwargs
        self.message: str = options.get("message", "")
        # self.filename=None
        # self.lineno=0
        # self.colno=0
        # self.error={}
        super().__init__(_type, options, *args, **kwargs)


class SubmitEvent(Event):
    """SubmitEvent"""

    SUBMIT: str = "submit"  #:

    def __init__(self, _type: str, options: dict = None, *args, **kwargs) -> None:
        options = options or kwargs  # if options is none use kwargs
        super().__init__(_type, options, *args, **kwargs)


class PointerEvent(Event):
    """PointerEvent"""

    POINTER: str = "pointer"  #:
    POINTERCANCEL: str = "pointercancel"  #:
    POINTERDOWN: str = "pointerdown"  #:
    POINTERENTER: str = "pointerenter"  #:
    POINTERLEAVE: str = "pointerleave"  #:
    POINTERMOVE: str = "pointermove"  #:
    POINTEROUT: str = "pointerout"  #:
    POINTEROVER: str = "pointerover"  #:
    POINTERUP: str = "pointerup"  #:

    def __init__(self, _type: str, options: dict = None, *args, **kwargs) -> None:
        options = options or kwargs  # if options is none use kwargs
        self.pointerId: float = options.get("pointerId", 0)
        self.width: float = options.get("width", 0)
        self.height: float = options.get("height", 0)
        self.pressure: float = options.get("pressure", 0)
        self.tangentialPressure: float = options.get("tangentialPressure", 0)
        self.tiltX: float = options.get("tiltX", 0)
        self.tiltY: float = options.get("tiltY", 0)
        self.twist: float = options.get("twist", 0)
        self.pointerType: str = options.get("pointerType", "")
        self.isPrimary: bool = options.get("isPrimary", False)
        super().__init__(_type, options, *args, **kwargs)

    def getCoalescedEvents(self):
        pass

    def getPredictedEvents(self):
        pass

    # def getCurrentPoint(self, element):
    #     """ Returns the current coordinates of the specified element relative to the viewport. """
    #     pass

    # def getIntermediatePoints(self, element):
    #     """ Returns the coordinates of all the intermediate points of the pointer along the path of the pointer. """
    #     pass


class BeforeUnloadEvent(Event):
    BEFOREUNLOAD = "beforeunload"  #:
    """ BeforeUnloadEvent """

    def __init__(self, _type: str, options: dict = None, *args, **kwargs) -> None:
        options = options or kwargs  # if options is none use kwargs
        super().__init__(_type, options, *args, **kwargs)


class SVGEvent(Event):
    """SVGEvent"""

    ABORT: str = "abort"  #:
    LOAD: str = "load"  #:
    LOADEDDATA: str = "loadeddata"  #:
    LOADEDMETADATA: str = "loadedmetadata"  #:
    LOADSTART: str = "loadstart"  #:
    PROGRESS: str = "progress"  #:
    SCROLL: str = "scroll"  #:
    UNLOAD: str = "unload"  #:
    ERROR: str = "error"  #:

    def __init__(self, _type: str, options: dict = None, *args, **kwargs) -> None:
        options = options or kwargs  # if options is none use kwargs
        super().__init__(_type, options, *args, **kwargs)

    # def initEvent(self, eventTypeArg, canBubbleArg, cancelableArg):
    #     pass


class TimerEvent(Event):
    TIMER: str = "timer"  #:
    TIMER_COMPLETE: str = "timercomplete"  #:
    """ TimerEvent """

    def __init__(self, _type: str, options: dict = None, *args, **kwargs) -> None:
        options = options or kwargs  # if options is none use kwargs
        super().__init__(_type, options, *args, **kwargs)

    # def initTimerEvent(self, type, bubbles, cancelable, detail):
    #     """ initTimerEvent() """
    #     pass


class DragEvent(Event):
    """DragEvent"""

    DRAG: str = "drag"  #:
    END: str = "dragend"  #:
    ENTER: str = "dragenter"  #:
    EXIT: str = "dragexit"  #:
    LEAVE: str = "dragleave"  #:
    OVER: str = "dragover"  #:
    START: str = "dragstart"  #:
    DROP: str = "drop"  #:

    def __init__(self, _type: str, options: dict = None, *args, **kwargs) -> None:
        options = options or kwargs  # if options is none use kwargs
        self.dataTransfer = options.get("dataTransfer", None)
        """ Returns the data that is dragged/dropped """
        super().__init__(_type, options, *args, **kwargs)


class HashChangeEvent(Event):
    """HashChangeEvent"""

    CHANGE: str = "hashchange"  #:

    def __init__(self, _type: str, options: dict = None, *args, **kwargs) -> None:
        options = options or kwargs  # if options is none use kwargs
        self.newURL = options.get("newURL", "")
        self.oldURL = options.get("oldURL", "")
        super().__init__(_type, options, *args, **kwargs)


class InputEvent(UIEvent):
    """InputEvent"""

    CHANGE: str = "change"  #:
    SELECT: str = "select"  #:
    INPUT: str = "input"  #:

    def __init__(self, _type: str, options: dict = None, *args, **kwargs) -> None:
        options = options or kwargs  # if options is none use kwargs
        self.data = options.get("data", None)
        """ Returns the inserted characters """
        self.dataTransfer = options.get("dataTransfer", None)
        """ Returns an object containing information about the inserted/deleted data """
        self.inputType = options.get("inputType", None)
        """ Returns the type of the change (i.e "inserting" or "deleting") """
        self.isComposing = options.get("isComposing", None)
        """ Returns whether the state of the event is composing or not """
        super().__init__(_type, options, *args, **kwargs)

    def getTargetRanges(self):
        """Returns an array containing target ranges that will be affected by the insertion/deletion"""
        pass


class PageTransitionEvent(Event):
    """PageTransitionEvent"""

    PAGEHIDE: str = "pagehide"  #:
    PAGESHOW: str = "pageshow"  #:

    def __init__(self, _type: str, options: dict = None, *args, **kwargs) -> None:
        options = options or kwargs  # if options is none use kwargs
        self.persisted = options.get("persisted", None)
        """ Returns whether the webpage was cached by the browser """
        super().__init__(_type, options, *args, **kwargs)


class PopStateEvent(Event):
    """PopStateEvent"""

    POPSTATE: str = "popstate"  #:
    # ONPOPSTATE = "onpopstate"  #:??

    def __init__(self, _type: str, options: dict = None, *args, **kwargs) -> None:
        options = options or kwargs  # if options is none use kwargs
        self.state = options.get("state", None)
        """ Returns an object containing a copy of the history entries """
        super().__init__(_type, options, *args, **kwargs)


class StorageEvent(Event):
    """StorageEvent"""

    STORAGE: str = "storage"  #:

    def __init__(self, _type: str, options: dict = None, *args, **kwargs) -> None:
        options = options or kwargs  # if options is none use kwargs
        self.key = options.get("key", None)
        """ Returns the key of the changed storage item """
        self.newValue = options.get("newValue", None)
        """ Returns the new value of the changed storage item """
        self.oldValue = options.get("oldValue", None)
        """ Returns the old value of the changed storage item """
        self.storageArea = options.get("storageArea", None)
        """ Returns an object representing the affected storage object """
        self.url = options.get("url", None)
        """ Returns the URL of the changed item's document """
        super().__init__(_type, options, *args, **kwargs)


class TransitionEvent(Event):
    """TransitionEvent"""

    TRANSITIONEND: str = "transitionend"  #:

    def __init__(self, _type: str, options: dict = None, *args, **kwargs) -> None:
        options = options or kwargs  # if options is none use kwargs
        self.propertyName = None
        """ Returns the name of the transition"""
        self.elapsedTime = None
        """  Returns the number of seconds a transition has been running """
        self.pseudoElement = None
        """ Returns the name of the pseudo-element of the transition """
        super().__init__(_type, options, *args, **kwargs)


class ProgressEvent(Event):
    """ProgressEvent"""

    LOADSTART: str = "loadstart"  #:
    PROGRESS: str = "progress"  #:
    ABORT: str = "abort"  #:
    ERROR: str = "error"  #:
    LOAD: str = "load"  #:
    LOADED: str = "loaded"  #:
    LOADEND: str = "loadend"  #:
    TIMEOUT: str = "timeout"  #:

    def __init__(self, _type: str, options: dict = None, *args, **kwargs) -> None:
        options = options or kwargs  # if options is none use kwargs
        self.lengthComputable: bool = options.get("lengthComputable", None)
        self.loaded: int = options.get("loaded", None)
        self.total: int = options.get("total", None)
        super().__init__(_type, options, *args, **kwargs)


class CustomEvent(Event):
    """CustomEvent"""

    def __init__(self, _type: str, options: dict = None, *args, **kwargs) -> None:
        options = options or kwargs  # if options is none use kwargs
        self.detail = None
        super().__init__(_type, options, *args, **kwargs)

    def initCustomEvent(self):
        pass


class GamePadEvent(Event):
    """GamePadEvent"""

    START: str = "gamepadconnected"  #:
    STOP: str = "gamepaddisconnected"  #:

    def __init__(self, _type: str, options: dict = None, *args, **kwargs) -> None:
        options = options or kwargs  # if options is none use kwargs
        self.gamepad = None
        super().__init__(_type, options, *args, **kwargs)


# TODO - tests and service worker API
class FetchEvent(Event):
    """FetchEvent"""

    FETCH: str = "fetch"  #:

    def __init__(self, _type: str, options: dict = None, *args, **kwargs) -> None:
        options = options or kwargs  # if options is none use kwargs
        self.clientId = None
        """ Returns the client ID of the fetch request """
        self.request = None
        """ Returns the request object """
        self.isReload = None
        """ Returns whether the request is a reload or not """
        super().__init__(_type, options, *args, **kwargs)

    @property
    def isReload(self):
        return self.request.url == self.request.referrer

    @property
    def replacesClientId(self):
        return self.clientId != self.request.clientId

    @property
    def resultingClientId(self):
        return self.clientId if self.replacesClientId else self.request.clientId

    def respondWith(self, response):
        """Returns a promise that resolves to the response object"""
        pass

    def waitUntil(self, promise):
        """Returns a promise that resolves when the response is available"""
        pass


class ExtendableEvent(Event):
    """ExtendableEvent"""

    # CAPTURING_PHASE = 1
    # AT_TARGET = 2
    # BUBBLING_PHASE = 3

    def __init__(self, _type: str, options: dict = None, *args, **kwargs) -> None:
        options = options or kwargs  # if options is none use kwargs
        self.extendable = None
        """ Returns whether the event is extendable or not """
        self.timeStamp = None
        """ Returns the time stamp of the event """
        # self.waitUntil(promise)
        """ Returns a promise that resolves when the event is handled """
        super().__init__(_type, options, *args, **kwargs)


class SyncEvent(ExtendableEvent):
    """SyncEvent"""

    SYNC: str = "sync"  #:

    def __init__(self, _type: str, options: dict = None, *args, **kwargs) -> None:
        options = options or kwargs  # if options is none use kwargs
        self.tag = None
        """ Returns the tag of the sync event """
        self.lastChance = None
        """ Returns whether the sync event is the last chance or not """
        super().__init__(_type, options, *args, **kwargs)


class SecurityPolicyViolationEvent(ExtendableEvent):
    """SecurityPolicyViolationEvent"""

    SECURITY_POLICY_VIOLATION: str = "securitypolicyviolation"  #:

    def __init__(self, _type: str, options: dict = None, *args, **kwargs) -> None:
        options = options or kwargs  # if options is none use kwargs
        self.blockedURI = None
        """ Returns the blocked URI """
        self.violatedDirective = None
        """ Returns the violated directive """
        self.originalPolicy = None
        """ Returns the original policy """
        self.isFrameAncestor = None
        """ Returns whether the frame is an ancestor of the frame that violated the policy """
        self.isMainFrame = None
        """ Returns whether the frame is the main frame """
        self.frame = None
        """ Returns the frame that violated the policy """
        super().__init__(_type, options, *args, **kwargs)


class DOMContentLoadedEvent(Event):
    """DOMContentLoadedEvent"""

    DOMCONTENTLOADED: str = "DOMContentLoaded"  #:
    # LOAD: str = "load"  #: already on event
    # BEFOREUNLOAD: str = "beforeunload"  #: already on event
    # UNLOAD: str = "unload"  #: already on event
    # readystatechange = "readystatechange"  #: ?? where does this one belong. Have added it to event

    def __init__(self, _type: str, options: dict = None, *args, **kwargs) -> None:
        options = options or kwargs  # if options is none use kwargs
        self.document = options.get("document", None)
        """ Returns the document that was loaded """
        super().__init__(_type, options, *args, **kwargs)


# class InstallEvent()

# class DeviceMotionEvent(Event):
#     """ DeviceMotionEvent """
#     def __init__(self, _type: str, options: dict = None, *args, **kwargs) -> None:
#         self.acceleration = None
#         """ Returns the acceleration of the device """
#         self.accelerationIncludingGravity = None
#         """ Returns the acceleration of the device, including gravity """
#         self.rotationRate = None
#         """ Returns the rotation rate of the device """
#         self.interval = None
#         """ Returns the time interval between the previous and the current event """
#         super().__init__(_type, options, *args, **kwargs)


# class DeviceOrientationEvent(Event):
#     """ DeviceOrientationEvent """
#     def __init__(self, _type: str, options: dict = None, *args, **kwargs) -> None:
#         self.absolute = None
#         """ Returns true if the orientation is absolute """
#         self.alpha = None
#         """ Returns the orientation of the device in degrees, relative to the Earth's coordinate system """
#         self.beta = None
#         """ Returns the orientation of the device in degrees, relative to the Earth's coordinate system """
#         self.gamma = None
#         """ Returns the orientation of the device in degrees, relative to the Earth's coordinate system """
#         self.interval = None
#         """ Returns the time interval between the previous and the current event """
#         super().__init__(_type, options, *args, **kwargs)


# class DeviceLightEvent(Event):
#     """ DeviceLightEvent """
#     def __init__(self, _type: str, options: dict = None, *args, **kwargs) -> None:
#         self.value = None
#         """ Returns the value of the ambient light sensor """
#         super().__init__(_type, options, *args, **kwargs)


# class DeviceProximityEvent(Event):
#     """ DeviceProximityEvent """
#     def __init__(self, _type: str, options: dict = None, *args, **kwargs) -> None:
#         self.value = None
#         """ Returns the value of the proximity sensor """
#         self.min = None
#         """ Returns the minimum value of the proximity sensor """
#         self.max = None
#         """ Returns the maximum value of the proximity sensor """
#         super().__init__(_type, options, *args, **kwargs)


class TweenEvent(Event):
    """TweenEvent"""

    START: str = "onStart"  #:
    STOP: str = "onStop"  #:
    RESET: str = "onReset"  #:
    PAUSE: str = "onPause"  #:
    UNPAUSE: str = "onUnPause"  #:
    UPDATE_START: str = "onUpdateStart"  #:
    UPDATE_END: str = "onUpdateEnd"  #:
    COMPLETE: str = "onComplete"  #:

    TIMER: str = "onTimer"  #:
    _source = None

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, source):
        self._source = source

    def __init__(self, _type, source=None, bubbles=False, cancelable=False):
        # super.__init__(self, type, bubbles, cancelable)
        super().__init__(_type)  # TODO -
        self.source = source


class PromiseRejectionEvent(Event):  # TODO - put with the promise?
    """PromiseRejectionEvent"""

    UNHANDLED: str = "unhandledrejection"  #:
    HANDLED: str = "rejectionhandled"  #:

    def __init__(self, _type, options=None, *args, **kwargs):
        self.promise = None
        """ Returns the promise that was rejected """
        self.reason = None
        """ Returns the reason of the rejection """
        self.isRejected = None
        """ Returns whether the promise was rejected or not """
        super().__init__(_type, options, *args, **kwargs)


class GlobalEventHandler:  # (EventDispatcher):

    # def __init__(self):
    #     super().__init__(self)
    #     self.addEventListener(KeyboardEvent.KEYDOWN, self.onkeydown)
    #     self.addEventListener(KeyboardEvent.KEYUP, self.onkeyup)

    #     self.addEventListener(MouseEvent.MOUSEMOVE, self.onmousemove)
    #     self.addEventListener(MouseEvent.MOUSEDOWN, self.onmousedown)
    #     self.addEventListener(MouseEvent.MOUSEUP, self.onmouseup)

    #     self.addEventListener(DragEvent.DRAG, self.ondrag)
    #     self.addEventListener(DragEvent.END, self.ondragend)
    #     self.addEventListener(DragEvent.ENTER, self.ondragenter)
    #     self.addEventListener(DragEvent.EXIT, self.ondragexit)
    #     self.addEventListener(DragEvent.LEAVE, self.ondragleave)
    #     self.addEventListener(DragEvent.OVER, self.ondragover)
    #     self.addEventListener(DragEvent.START, self.ondragstart)
    #     self.addEventListener(DragEvent.DROP, self.ondrop)

    #     self.addEventListener(ClipboardEvent.CUT, self.oncut)
    #     self.addEventListener(ClipboardEvent.COPY, self.oncopy)
    #     self.addEventListener(ClipboardEvent.PASTE, self.onpaste)

    def onabort(self, event):
        print(event)
        raise NotImplementedError

    def onblur(self, event):
        print(event)
        raise NotImplementedError

    def oncancel(self, event):
        print(event)
        raise NotImplementedError

    def oncanplay(self, event):
        print(event)
        raise NotImplementedError

    def oncanplaythrough(self, event):
        print(event)
        raise NotImplementedError

    def onchange(self, event):
        print(event)
        raise NotImplementedError

    def onclick(self, event):
        print(event)
        raise NotImplementedError

    def onclose(self, event):
        print(event)
        raise NotImplementedError

    def oncontextmenu(self, event):
        print(event)
        raise NotImplementedError

    def oncuechange(self, event):
        print(event)
        raise NotImplementedError

    def ondblclick(self, event):
        print(event)
        raise NotImplementedError

    def ondrag(self, event):
        print(event)
        raise NotImplementedError

    def ondragend(self, event):
        print(event)
        raise NotImplementedError

    def ondragenter(self, event):
        print(event)
        raise NotImplementedError

    def ondragexit(self, event):
        print(event)
        raise NotImplementedError

    def ondragleave(self, event):
        print(event)
        raise NotImplementedError

    def ondragover(self, event):
        print(event)
        raise NotImplementedError

    def ondragstart(self, event):
        print(event)
        raise NotImplementedError

    def ondrop(self, event):
        print(event)
        raise NotImplementedError

    def ondurationchange(self, event):
        print(event)
        raise NotImplementedError

    def onemptied(self, event):
        print(event)
        raise NotImplementedError

    def onended(self, event):
        print(event)
        raise NotImplementedError

    def onerror(self, event):
        print(event)
        raise NotImplementedError

    def onfocus(self, event):
        print(event)
        raise NotImplementedError

    def ongotpointercapture(self, event):
        print(event)
        raise NotImplementedError

    def oninput(self, event):
        print(event)
        raise NotImplementedError

    def oninvalid(self, event):
        print(event)
        raise NotImplementedError

    def onkeydown(self, event):
        print(event)
        raise NotImplementedError

    def onkeypress(self, event):
        print(event)
        raise NotImplementedError

    def onkeyup(self, event):
        print(event)
        raise NotImplementedError

    def onload(self, event):
        print(event)
        raise NotImplementedError

    def onloadeddata(self, event):
        print(event)
        raise NotImplementedError

    def onloadedmetadata(self, event):
        print(event)
        raise NotImplementedError

    def onloadend(self, event):
        print(event)
        raise NotImplementedError

    def onloadstart(self, event):
        print(event)
        raise NotImplementedError

    def onlostpointercapture(self, event):
        print(event)
        raise NotImplementedError

    def onmouseenter(self, event):
        print(event)
        raise NotImplementedError

    def onmouseleave(self, event):
        print(event)
        raise NotImplementedError

    def onmousemove(self, event):
        print(event)
        raise NotImplementedError

    def onmouseout(self, event):
        print(event)
        raise NotImplementedError

    def onmouseover(self, event):
        print(event)
        raise NotImplementedError

    def onmouseup(self, event):
        print(event)
        raise NotImplementedError

    def onpause(self, event):
        print(event)
        raise NotImplementedError

    def onplay(self, event):
        print(event)
        raise NotImplementedError

    def onplaying(self, event):
        print(event)
        raise NotImplementedError

    def onpointercancel(self, event):
        print(event)
        raise NotImplementedError

    def onpointerdown(self, event):
        print(event)
        raise NotImplementedError

    def onpointerenter(self, event):
        print(event)
        raise NotImplementedError

    def onpointerleave(self, event):
        print(event)
        raise NotImplementedError

    def onpointermove(self, event):
        print(event)
        raise NotImplementedError

    def onpointerout(self, event):
        print(event)
        raise NotImplementedError

    def onpointerover(self, event):
        print(event)
        raise NotImplementedError

    def onpointerup(self, event):
        print(event)
        raise NotImplementedError

    def onprogress(self, event):
        print(event)
        raise NotImplementedError

    def onratechange(self, event):
        print(event)
        raise NotImplementedError

    def onreset(self, event):
        print(event)
        raise NotImplementedError

    def onresize(self, event):
        print(event)
        raise NotImplementedError

    def onscroll(self, event):
        print(event)
        raise NotImplementedError

    def onseeked(self, event):
        print(event)
        raise NotImplementedError

    def onseeking(self, event):
        print(event)
        raise NotImplementedError

    def onselect(self, event):
        print(event)
        raise NotImplementedError

    def onselectionchange(self, event):
        print(event)
        raise NotImplementedError

    def onselectstart(self, event):
        print(event)
        raise NotImplementedError

    def onshow(self, event):
        print(event)
        raise NotImplementedError

    def onstalled(self, event):
        print(event)
        raise NotImplementedError

    def onsubmit(self, event):
        print(event)
        raise NotImplementedError

    def onsuspend(self, event):
        print(event)
        raise NotImplementedError

    def ontimeupdate(self, event):
        print(event)
        raise NotImplementedError

    def onvolumechange(self, event):
        print(event)
        raise NotImplementedError

    def onwaiting(self, event):
        print(event)
        raise NotImplementedError

    def onwheel(self, event):
        print(event)
        raise NotImplementedError

    def onanimationcancel(self, event):
        print(event)
        raise NotImplementedError

    def onanimationend(self, event):
        print(event)
        raise NotImplementedError

    def onanimationiteration(self, event):
        print(event)
        raise NotImplementedError

    def onauxclick(self, event):
        print(event)
        raise NotImplementedError

    def onformdata(self, event):
        print(event)
        raise NotImplementedError

    def onmousedown(self, event):
        print(event)
        raise NotImplementedError

    def ontouchcancel(self, event):
        print(event)
        raise NotImplementedError

    def ontouchstart(self, event):
        print(event)
        raise NotImplementedError

    def ontransitioncancel(self, event):
        print(event)
        raise NotImplementedError

    def ontransitionend(self, event):
        print(event)
        raise NotImplementedError


class WindowEventHandler:  # (EventHandler): # TODO - put in the window module?
    def __init__(self, window):
        super().__init__()
        self.window = window

    def onabort(self, event):
        print(event)
        raise NotImplementedError

    def onafterprint(self, event):
        print(event)
        raise NotImplementedError

    def onbeforeprint(self, event):
        print(event)
        raise NotImplementedError

    def onbeforeunload(self, event):
        print(event)
        raise NotImplementedError

    def onblur(self, event):
        print(event)
        raise NotImplementedError

    def oncanplay(self, event):
        print(event)
        raise NotImplementedError

    def oncanplaythrough(self, event):
        print(event)
        raise NotImplementedError

    def onchange(self, event):
        print(event)
        raise NotImplementedError

    def onclick(self, event):
        print(event)
        raise NotImplementedError

    def oncontextmenu(self, event):
        print(event)
        raise NotImplementedError

    def oncopy(self, event):
        print(event)
        raise NotImplementedError

    def oncuechange(self, event):
        print(event)
        raise NotImplementedError

    def oncut(self, event):
        print(event)
        raise NotImplementedError

    def ondblclick(self, event):
        print(event)
        raise NotImplementedError

    def ondrag(self, event):
        print(event)
        raise NotImplementedError

    def ondragend(self, event):
        print(event)
        raise NotImplementedError

    def ondragenter(self, event):
        print(event)
        raise NotImplementedError

    def ondragleave(self, event):
        print(event)
        raise NotImplementedError

    def ondragover(self, event):
        print(event)
        raise NotImplementedError



"""
    domonic.geom
    ====================================

    much of this was written by.ai
    (https://6b.eleuther.ai/)

    most of this code was derivied by continually pasting back into eleuther

    # TODO - tests for all these new functions. sadly the ai didn't generate those.

"""
import math
# import operator


class vec2():
    """[vec2]
    """

    def __init__(self, x: float = 0, y: float = 0):
        self.x: float = x
        self.y: float = y

    def __mul__(self, other):
        """[vec2] * [number] or [vec2] * [vec2]]

        Args:
            other ([vec2]): [multiply this vector by this vector or by a number]

        Returns:
            [vec2]: [new vector]
        """
        if isinstance(other, vec2):
            return self.__class__((self.x * other.x, self.y * other.y))
        else:
            return self.__class__((self.x * other, self.y * other))

    def __div__(self, other):
        if isinstance(other, vec2):
            return self.__class__((self.x / other.x, self.y / other.y))
        else:
            return self.__class__((self.x / other, self.y / other))

    def __mod__(self, other):
        if isinstance(other, vec2):
            return self.__class__(self.x % other.x, self.y % other.y)
        else:
            return self.__class__(self.x % other, self.y % other)

    def __getitem__(self, item):
        if isinstance(item, int):
            if item == 0:
                return self.x
            elif item == 1:
                return self.y
        elif isinstance(item, str):
            if item == 'x':
                return self.x
            elif item == 'y':
                return self.y

    def __len__(self):
        return 2

    def __add__(self, other):
        if isinstance(other, vec2):
            return vec2(self.x + other.x, self.y + other.y)
        else:
            return vec2(self.x + other, self.y + other)

    def __sub__(self, other):
        if isinstance(other, vec2):
            return vec2(self.x - other.x, self.y - other.y)
        else:
            return vec2(self.x - other, self.y - other)

    def __truediv__(self, other):
        if isinstance(other, vec2):
            return vec2(self.x / other.x, self.y / other.y)
        else:
            return vec2(self.x / other, self.y / other)

    def __floordiv__(self, other):
        if isinstance(other, vec2):
            return vec2(self.x // other.x, self.y // other.y)
        else:
            return vec2(self.x // other, self.y // other)

    def __neg__(self):
        return vec2(-self.x, -self.y)

    def __eq__(self, other):
        return (self.x == other.x and self.y == other.y)

    def __ne__(self, other):
        return (self.x != other.x or self.y != other.y)

    def __lt__(self, other):
        return (self.x < other.x and self.y < other.y)

    def __le__(self, other):
        return (self.x <= other.x and self.y <= other.y)

    def __gt__(self, other):
        return (self.x > other.x and self.y > other.y)

    def __ge__(self, other):
        return (self.x >= other.x and self.y >= other.y)

    def __hash__(self):
        return hash((self.x, self.y))

    def __setitem__(self, key, value):
        if key == 0:
            self.x = value
        elif key == 1:
            self.y = value
        else:
            raise IndexError("Index out of range")

    def __iter__(self):
        yield self.x
        yield self.y

    def __call__(self, *args, **kwargs):
        return self.x, self.y

    def __getstate__(self):
        return self.x, self.y

    def __copy__(self):
        return vec2(self.x, self.y)

    def __deepcopy__(self, memo):
        return vec2(self.x, self.y)

    def __round__(self, n):
        return vec2(round(self.x, n), round(self.y, n))

    def __str__(self):
        return str(self.x) + " " + str(self.y)

    def add(self, pt):
        """ Adds a point to this point. returns new one """
        return vec2(self.x + pt.x, self.y + pt.y)

    def subtract(self, pt):
        """ Subtract from this point. returns new one """
        return vec2(self.x - pt.x, self.y - pt.y)

    def copy(self):
        """ Creates a copy of this Point object. """
        return vec2(self.x, self.y)

    def dot(self, other):
        """ Returns the dot product of these two vectors """
        x1 = self.x
        y1 = self.y
        x2 = other.x
        y2 = other.y
        return x1 * x2 + y1 * y2

    def cross(self, other):
        """ Returns the cross product of this and other. The cross product is considered 0
        perpendicular to each other point """
        x1 = self.x
        y1 = self.y
        x2 = other.x
        y2 = other.y
        return x1 * y2 - y1 * x2

    def negative(self):
        """ return vector but negative version """
        return vec2(-self.x, -self.y)

    def distance(self, other):
        x_dot = self.x - other.x
        y_dot = self.y - other.y
        return math.sqrt(x_dot * x_dot + y_dot * y_dot)

    def squareDistance(self, other):
        return self.distance(self) * self.distance(other)

    def equals(self, other):
        """ Returns true if all vectors components are the same """
        return (self.x == other.x and self.y == other.y)

    def interpolate(self, pt1, pt2, t):
        """ Calculates the point that would be reached by this Point
        if we moved by a given distance vector over time t.
        Only floats are handled here. """
        return vec2(self.x * (1 - t) + pt1.x * t, self.y * (1 - t) + pt1.y * t)

    def copyRotateAround(self, target, angle):
        """moves the vector then rotate"""
        self.rotateAround(target, angle)
        return vec2(self.x, self.y)

    def mirrorHorizontally(self):
        return self.rotateAround(vec2(1, 0), 90)

    def mirrorVertically(self):
        return self.rotateAround(vec2(0, 1), 90)

    def negate(self):
        """Negate a vector. This results in vector with the same direction but different length"""
        return vec2(-self.x, -self.y)

    def length(self):
        """ returns the length of this vector """
        return self.x

    def squaredlength(self):
        """ returns the squared length of this vector """
        return self.x * self.x + self.y * self.y

    def normalize(self):
        """ returns a normalized vector """
        return self / self.length()

    def obj(self):
        """ returns a obj representation of this vector """
        return {"x": self.x, "y": self.y}

    # def __repr__(self):
    #     return str(self.x) + " " + str(self.y)

    def json(self):
        """ returns a json representation of this vector """
        return str({"x": self.x, "y": self.y})

    def angle(self):
        """ returns the angle of this vector in radians """
        return math.atan2(self.y, self.x)

    def angleBetween(self, other):
        """ returns the angle between this and another vector in radians """
        return math.acos(self.dot(other) / (self.length() * other.length()))

    def rotate(self, angle):
        """ rotates this vector by an angle in radians """
        x = self.x
        y = self.y
        self.x = x * math.cos(angle) - y * math.sin(angle)
        self.y = y * math.cos(angle) + x * math.sin(angle)

    def rotateAround(self, target, angle):
        """Rotates the vector around another point. In fact it returns the other point."""
        dot = self.dot(target)
        s = math.sin(angle / 2)
        c = math.cos(angle / 2)
        x = s * (self.x - dot * c) + c
        y = s * (self.y - dot * c) - c
        return vec2(x, y)

    @staticmethod
    def cmp(a, b):
        """ Compare two vectors. """
        return (a > b) - (a < b)

    # @staticmethod
    # def random():
    #     """ returns a random vector """
    #     return vec2(random.random(), random.random())

    @staticmethod
    def random(min_x, max_x, min_y, max_y):
        """ returns a random vector """
        import random
        return vec2(random.uniform(min_x, max_x), random.uniform(min_y, max_y))

    @staticmethod
    def random_unit_vector(min_x, max_x, min_y, max_y):
        """ returns a random unit vector """
        vec = vec2.random(min_x, max_x, min_y, max_y)
        return vec.normalize()

    @staticmethod
    def random_point_in_circle(center, radius):
        """ returns a random point in a circle """
        vec = vec2.random(0, radius)
        return vec.add(center)

    @staticmethod
    def random_point_in_rectangle(min_x, max_x, min_y, max_y):
        """ returns a random point in a rectangle """
        vec = vec2.random(min_x, max_x, min_y, max_y)
        return vec.add(vec2.random(0, 1))

    # def rotateAround(self, target, angle):
    #     """ Rotates the vector around another point. In fact it returns the other point."""
    #     dot = self.dot(target)
    #     s = math.sin(angle)
    #     c = math.cos(angle)
    #     x = s * (self.x - dot * c) + c
    #     y = s * (self.y - dot * c) - c
    #     return vec2(x, y)

    # def rotateAroundLocal(self, point, angle):
    #     """ Rotates the vector around a point. In fact it returns itself."""
    #     dot = self.dot(point)
    #     s = math.sin(angle)
    #     c = math.cos(angle)
    #     x = s * (self.x - dot * c) + c
    #     y = s * (self.y - dot * c) - c
    #     return vec2(x, y)

    # def rotateLocal(self, point, angle):
    #     """ Rotates the vector around a point. In fact it returns itself."""
    #     dot = self.dot(point)
    #     s = math.sin(angle)
    #     c = math.cos(angle)
    #     x = s * (self.x - dot * c) + c
    #     y = s * (self.y - dot * c) - c
    #     self.x = x
    #     self.y = y
    #     return self



"""
    test_dQuery
    ~~~~~~~~~~~~~~~
    unit tests for domonic.dQuery

"""

import time
import unittest
# import requests
# from mock import patch
# from domonic.javascript import Math

from domonic.dom import *
from domonic.html import *
from domonic.dQuery import *


class TestCase(unittest.TestCase):

    # domonic.dQuery.º
    def test_hello(self):
        d = html(head(body(li(_class='things'), div(_id="test"))))
        º(d)
        print('---** -')
        print(º('#test'))
        print('---** -')
        print(º('.things'))
        print('---** -')

        print('a::')
        a = º('<div class="test2"></div>')
        print(a)

        print('b::')
        print(º('#test'))
        b = º('#test').append(a)
        print(b)

        print(d)

        pass

    def test_add(self):
        # test = º('<p></p>').add('<h1>').add(div())
        # print(test)
        pass

    def test_addBack(self):
        pass

    def test_addClass(self):
        a = º('<div id="test2"></div><div id="test3"></div>')
        assert str(a) == '<div id="test2"></div><div id="test3"></div>'
        a.addClass('one')
        assert str(a) == '<div id="test2" class="one"></div><div id="test3" class="one"></div>'
        # print("1:",a)
        # print("2:",str(a))
        # print(str(a))
        a.addClass('one').addClass('two').addClass('three')
        assert str(a) == '<div id="test2" class="one one two three"></div><div id="test3" class="one one two three"></div>'
        # for el in a.elements:
        # print(el.getAttribute("class"))

    def test_after(self):
        # TODO - sort the parser... positional error on this as not multiline
        # tags = º('<div id="test1"><h1>asd</h1></div>')
        # print(tags)
        app = html(head(), body(div(span(), _id="test")))
        º(app)  # TODO _str is none?
        # print( 'wtf:??:', º('#test1') ) # TODO - better errors when passing wrong id name
        º('#test').after(p('hi'))
        print(app)
        # pass

    def test_ajaxComplete(self):
        pass

    def test_ajaxError(self):
        pass

    def test_ajaxSend(self):
        pass

    def test_ajaxStart(self):
        pass

    def test_ajaxStop(self):
        pass

    def test_ajaxSuccess(self):
        pass

    def test_andSelf(self):
        pass

    def test_animate(self):
        pass

    def test_append(self):
        print("TEST APPEND")
        # doc = html()
        # º(doc)
        # doc.append("some text")
        # print(doc.html())
        # d = º('<div></div>').append("some text")
        # self.assertEqual(str(d), '<div>some text</div>')
        pass

    def test_appendTo(self):
        pass

    def test_attr(self):
        a = º('<div id="test2"></div>')
        a.addClass('one')
        assert str(a) == '<div id="test2" class="one"></div>'
        assert a.attr('id') == 'test2'
        assert a.attr('class') == 'one'
        a.attr('id', 'somethingelse')
        assert str(a) == '<div id="somethingelse" class="one"></div>'
        # print(a.elements[0])

    def test_before(self):
        pass

    def test_bind(self):
        pass

    def test_blur(self):
        pass

    def test_change(self):
        pass

    def test_children(self):
        pass

    def test_clearQueue(self):
        pass

    def test_click(self):
        pass

    def test_clone(self):
        pass

    def test_closest(self):
        pass

    def test_contents(self):
        pass

    def test_context(self):
        pass

    def test_contextmenu(self):
        pass

    def test_css(self):
        pass

    def test_data(self):
        pass

    def test_dblclick(self):
        pass

    def test_delay(self):
        pass

    def test_delegate(self):
        pass

    def test_dequeue(self):
        pass

    def test_detach(self):
        pass

    def test_die(self):
        pass

    def test_each(self):
        pass

    def test_empty(self):
        pass

    def test_end(self):
        pass

    def test_eq(self):
        pass

    def test_error(self):
        pass

    def test_even(self):
        pass

    def test_fadeIn(self):
        pass

    def test_fadeOut(self):
        pass

    def test_fadeTo(self):
        pass

    def test_fadeToggle(self):
        pass

    def test_filter(self):
        pass

    def test_find(self):
        pass

    def test_finish(self):
        pass

    def test_first(self):
        pass

    def test_focus(self):
        pass

    def test_focusin(self):
        pass

    def test_focusout(self):
        pass

    def test_get(self):
        pass

    def test_has(self):
        pass

    def test_hasClass(self):
        a = º('<div id="test2"></div>')
        a.addClass('one').addClass('two').addClass('three')
        assert a.hasClass('one') == True
        assert a.hasClass('five') == False

    def test_height(self):
        pass

    def test_hide(self):
        pass

    def test_hover(self):
        pass

    def test_html(self):
        pass

    def test_index(self):
        pass

    def test_innerHeight(self):
        pass

    def test_innerWidth(self):
        pass

    def test_insertAfter(self):
        pass

    def test_insertBefore(self):
        pass

    # def test_is(self):
        # pass

    def test_keydown(self):
        pass

    def test_keypress(self):
        pass

    def test_keyup(self):
        pass

    def test_last(self):
        things = º('<li></li><li></li><li></li><li></li><li data-tag="me"></li>')
        assert str(things.last()) == '<li data-tag="me"></li>'

    def test_length(self):
        pass

    def test_live(self):
        pass

    def test_load(self):
        pass

    def test_map(self):
        pass

    def test_mousedown(self):
        pass

    def test_mouseenter(self):
        pass

    def test_mouseleave(self):
        pass

    def test_mousemove(self):
        pass

    def test_mouseout(self):
        pass

    def test_mouseover(self):
        pass

    def test_mouseup(self):
        pass

    def test_next(self):
        pass

    def test_nextAll(self):
        pass

    def test_nextUntil(self):
        pass

    # def test_not(self):
        # pass

    def test_odd(self):
        pass

    def off(self, event):
        pass

    def test_offset(self):
        pass

    def test_offsetParent(self):
        pass

    def on(self, event, callback):
        pass

    def test_one(self):
        pass

    def test_outerHeight(self):
        pass

    def test_outerWidth(self):
        pass

    def test_parent(self):
        pass

    def test_parents(self):
        pass

    def test_parentsUntil(self):
        pass

    def test_position(self):
        pass

    def prepend(self, html):
        pass

    def test_prependTo(self):
        pass

    def test_prev(self):
        pass

    def test_prevAll(self):
        pass

    def test_prevUntil(self):
        pass

    def test_promise(self):
        pass

    def test_prop(self):
        pass

    def test_pushStack(self):
        pass

    def test_queue(self):
        pass

    def test_ready(self):
        pass

    def test_remove(self):
        pass

    def test_removeAttr(self):
        pass

    def test_removeClass(self):
        a = º('<div id="test2"></div>')
        a.addClass('one').addClass('two').addClass('three')
        assert a.hasClass('one') == True
        a.removeClass('one')
        assert a.hasClass('one') == False

    def test_removeData(self):
        pass

    def test_removeProp(self):
        pass

    def test_replaceAll(self):
        pass

    def test_replaceWith(self):
        pass

    def test_resize(self):
        pass

    def test_scroll(self):
        pass

    def test_scrollLeft(self):
        pass

    def test_scrollTop(self):
        pass

    def test_select(self):
        pass

    def test_serialize(self):
        page = html(form(
            select(_name="single",).html(
                option("Single", _selected=True),
                option("Single2")
            ),
            br(),
            select(_name="multiple", _multiple="multiple").html(
                option("Multiple", _selected="selected"),
                option("Multiple2"),
                option("Multiple3", _selected="selected")
            ),
            input(_type="text", _id="lname", _name="lname")
        ))
        º(page)
        print(º('form').serialize())

    def test_serializeArray(self):
        pass

    def test_show(self):
        pass

    def test_siblings(self):
        pass

    def test_size(self):
        pass

    def test_slice(self):
        pass

    def test_slideDown(self):
        pass

    def test_slideToggle(self):
        pass

    def test_slideUp(self):
        pass

    def test_stop(self):
        pass

    def test_submit(self):
        pass

    def test_text(self):
        page = html(form(
                select(_name="single",).html(
                    option("a", _selected=True),
                    option("b")
                ),
            ),
            div('hi'),
            div(span('there'))
        )
        º(page)
        assert º('div').text() == ['hi', 'there']
        º('div').text('test')
        assert º('div').text() == ['test', 'test']
        assert str(page) == '<html><form><select name="single"><option selected="true">a</option><option>b</option></select></form><div>test</div><div>test</div></html>'

    def test_toArray(self):
        pass

    def test_toggle(self):
        pass

    def test_toggleClass(self):
        # page = html(form(
        #         select(_name="single",).html(
        #             option("a", _selected=True),
        #             option("b")
        #         ), _id='test'
        #     ),
        #     div('hi'),
        #     div(span('there'))
        # )
        # º(page)
        # print(page)
        # º('#test').toggleClass('someclass')
        # print(page)
        # º('#test').toggleClass('someclass')
        # print(page)
        pass

    def test_trigger(self):
        pass

    def test_triggerHandler(self):
        pass

    def test_unbind(self):
        pass

    def test_undelegate(self):
        pass

    def test_unload(self):
        pass

    def test_unwrap(self):
        pass

    def test_val(self):
        pass

    def test_width(self):
        pass

    def test_wrap(self):
        pass

    def test_wrapAll(self):
        pass

    def test_wrapInner(self):
        pass

    def test_staticmethods(self):
        print("test_staticmethods::::::::::::::::::")

        d = html()
        º(d)

        d.appendChild(body())

        # º.boxModel
        # º.browser
        # º.cssHooks
        # º.cssNumber
        # º.ready
        # º.speed
        # º.support

        # º.ajax()
        # º.ajaxPrefilter()
        # º.ajaxSetup()
        # º.ajaxTransport()
        # º.Callbacks()

        # print(d)
        # print('el:', d.documentElement)
        # print('bod:', d.body)
        assert º.contains(d.documentElement, d.body) == True  # true
        assert º.contains(d.body, d.documentElement) == False  # false

        # º.data()
        # º.Deferred()
        # º.dequeue()
        # º.each()
        # º.error()
        # º.escapeSelector()

        obj1 = {'a': 1, 'b': 2}
        obj2 = {'c': 1, 'b': 5}
        print(º.extend(obj1, obj2))

        test = lambda x: x
        test2 = 1
        print("well?:", º.isFunction(test))
        print("well?:", º.isFunction(test2))

        # º.get()
        # º.getJSON()
        # º.getScript()
        # º.globalEval()
        # º.grep()
        # º.hasData()
        # º.holdReady()
        # º.htmlPrefilter()
        # º.inArray()
        # º.isArray()
        # º.isEmptyObject()
        # º.isNumeric()
        # º.isPlainObject()
        # º.isWindow()
        # º.isXMLDoc()
        # º.makeArray()
        # º.map()

        first = ["a", "b", "c"]
        second = ["d", "e", "f"]
        result = º.merge(º.merge([], first), second)
        print(first)
        print(second)
        print(result)

        first = ["a", "b", "c"]
        second = ["d", "e", "f"]
        result = º.merge(first, second)
        print(first)
        print(second)
        print(result)

        # º.noConflict()
        # º.noop()

        print(º.now())

        # º.param()
        # º.parseHTML()
        # º.parseJSON()
        # º.parseXML()
        # º.post()
        # º.proxy()
        # º.queue()
        # º.readyException()
        # º.removeData()
        # º.sub()
        print(º.trim("  some tst \n   TEST."))
        # º.type()
        # º.unique()
        # º.uniqueSort()
        # º.when()


if __name__ == '__main__':
    unittest.main()



"""
    domonic.i18n.da
    ====================================

    html tags in Danish

    first translated to Danish using google translate.
    send a pull request if you have a better translation.
"""
from domonic.html import *

krop = body
hovedet = head
# script = script
stil = style
# portal = portal
stærk = strong
blokcitat = blockquote
tabel = table
# form = form
etiket = label
indsend = submit
titel = title
# noscript = noscript
afsnit = section
# nav = nav
artikel = article
tilside = aside
hgruppe = hgroup
adresse = address
figur = figure
figcaption = figcaption
lille = small
citere = cite
kode = code
# sub = sub
# sup = sup
mærke = mark
rubin = ruby
# span = span
# iframe = iframe
# video = video
lyd = audio
lærred = canvas
billedtekst = caption
colgroup = colgroup
krop = tbody
# thead = thead
tfod = tfoot
feltsæt = fieldset
legende = legend
knap = button
vælg = select
dataliste = datalist
# optgroup = optgroup
mulighed = option
tekstområde = textarea
# output = output
fremskridt = progress
# meter = meter
detaljer = details
resumé = summary
# menu = menu
# menuitem = menuitem
skrifttype = font
overskrift = header
sidefod = footer
# base = base
# link = link
# meta = meta
# img = img
indlejre = embed
kilde = source
spor = track
område = area
# col = col
# input = input
kommando = command
hoved = main
# basefont = basefont
centrum = center
indlejre = embed
skrifttype = font
isindeks = isindex
liste = listing
klartekst = plaintext
strejke = strike
skabelon = template
billede = picture
# doctype = doctype
kommentar = comment



"""
    domonic.i18n.sw
    ====================================

    html tags in Swahili

    translated to Swahili using google translate.
    send a pull request if you have a better translation.

    html(
        kichwa(),
        mwili(
            div('Salamu, Dunia')
        )
    )
"""
from domonic.html import *

# tags
mwili = body
kichwa = head
hati = script
mtindo = style
# portal = portal
nguvu = strong
# blockquote = blockquote
meza = table
fomu = form
lebo = label
kuwasilisha = submit
# kichwa = title
# noscript = noscript
sehemu = section
# nav = nav
makala = article
pembeni = aside
kundi = hgroup
anwani = address
takwimu = figure
# figcaption = figcaption
ndogo = small
taja = cite
kanuni = code
ndogo = sub
chakula = sup
alama = mark
# ruby = ruby
muda = span
# iframe = iframe
# video = video
sauti = audio
turubai = canvas
maelezo = caption
kikundi = colgroup
# mwili = tbody
# kichwa = thead
mguu = tfoot
uwanja = fieldset
hadithi = legend
kitufe = button
chagua = select
orodhayadata = datalist
optgroup = optgroup
chaguo = option
eneolamaandishi = textarea
pato = output
maendeleo = progress
mita = meter
maelezo = details
muhtasari = summary
menyu = menu
# menyu = menuitem
fonti = font
# kichwa = header
chini = footer
msingi = base
kiungo = link
# meta = meta
# img = img
# embed = embed
chanzo = source
wimbo = track
eneo = area
col = col
pembejeo = input
amri = command
kuu = main
fonti = basefont
kituo = center
# embed = embed
fonti = font
# isindex = isindex
orodha = listing
maandishiwazi = plaintext
mgomo = strike
# template = template
picha = picture
# doctype = doctype
maoni = comment



"""
    domonic.aframe
    ====================================
    Generate aframe tags with python 3

    https://aframe.io/

"""

from domonic.html import tag  # , closed_tag
from domonic.dom import Node, ParentNode


class aframe_tag(tag):
    def __str__(self):
        return f"<a-{self.name}{self.__attributes__}>{self.content}</a-{self.name}>"


def aframe_init(self, *args, **kwargs):
    tag.__init__(self, *args, **kwargs)
    Node.__init__(self, *args, **kwargs)


entity = type('entity', (aframe_tag, Node, ParentNode), {'name': 'entity', '__init__': aframe_init})
scene = type('scene', (aframe_tag, Node, ParentNode), {'name': 'scene', '__init__': aframe_init})
material = type('material', (aframe_tag, Node, ParentNode), {'name': 'material', '__init__': aframe_init})
appearance = type('appearance', (aframe_tag, Node, ParentNode), {'name': 'appearance', '__init__': aframe_init})

shape = type('shape', (aframe_tag, Node, ParentNode), {'name': 'shape', '__init__': aframe_init})
# transform = type('transform', (aframe_tag, Node, ParentNode), {'name': 'transform', '__init__': aframe_init})
# inline = type('inline', (aframe_tag, Node, ParentNode), {'name': 'inline', '__init__': aframe_init})

sphere = type('sphere', (aframe_tag, Node, ParentNode), {'name': 'sphere', '__init__': aframe_init})
box = type('box', (aframe_tag, Node, ParentNode), {'name': 'box', '__init__': aframe_init})
plane = type('plane', (aframe_tag, Node, ParentNode), {'name': 'plane', '__init__': aframe_init})
sky = type('sky', (aframe_tag, Node, ParentNode), {'name': 'sky', '__init__': aframe_init})

mixin = type('mixin', (aframe_tag, Node, ParentNode), {'name': 'mixin', '__init__': aframe_init})
circle = type('circle', (aframe_tag, Node, ParentNode), {'name': 'circle', '__init__': aframe_init})
camera = type('camera', (aframe_tag, Node, ParentNode), {'name': 'camera', '__init__': aframe_init})
cone = type('cone', (aframe_tag, Node, ParentNode), {'name': 'cone', '__init__': aframe_init})
cursor = type('cursor', (aframe_tag, Node, ParentNode), {'name': 'cursor', '__init__': aframe_init})

curvedimage = type('curvedimage', (aframe_tag, Node, ParentNode), {'name': 'curvedimage', '__init__': aframe_init})
cylinder = type('cylinder', (aframe_tag, Node, ParentNode), {'name': 'cylinder', '__init__': aframe_init})

dodecahedron = type('dodecahedron', (aframe_tag, Node, ParentNode), {'name': 'dodecahedron', '__init__': aframe_init})
icosahedron = type('icosahedron', (aframe_tag, Node, ParentNode), {'name': 'icosahedron', '__init__': aframe_init})
image = type('image', (aframe_tag, Node, ParentNode), {'name': 'image', '__init__': aframe_init})
light = type('light', (aframe_tag, Node, ParentNode), {'name': 'light', '__init__': aframe_init})
alink = type('alink', (aframe_tag, Node, ParentNode), {'name': 'alink', '__init__': aframe_init})
octahedron = type('octahedron', (aframe_tag, Node, ParentNode), {'name': 'octahedron', '__init__': aframe_init})
plane = type('plane', (aframe_tag, Node, ParentNode), {'name': 'plane', '__init__': aframe_init})
ring = type('ring', (aframe_tag, Node, ParentNode), {'name': 'ring', '__init__': aframe_init})
sound = type('sound', (aframe_tag, Node, ParentNode), {'name': 'sound', '__init__': aframe_init})
sphere = type('sphere', (aframe_tag, Node, ParentNode), {'name': 'sphere', '__init__': aframe_init})
tetrahedron = type('tetrahedron', (aframe_tag, Node, ParentNode), {'name': 'tetrahedron', '__init__': aframe_init})
text = type('text', (aframe_tag, Node, ParentNode), {'name': 'text', '__init__': aframe_init})
torus = type('torus', (aframe_tag, Node, ParentNode), {'name': 'torus', '__init__': aframe_init})
triangle = type('triangle', (aframe_tag, Node, ParentNode), {'name': 'triangle', '__init__': aframe_init})
video = type('video', (aframe_tag, Node, ParentNode), {'name': 'video', '__init__': aframe_init})
videosphere = type('videosphere', (aframe_tag, Node, ParentNode), {'name': 'videosphere', '__init__': aframe_init})

gltf = type('gltf', (aframe_tag, Node, ParentNode), {'name': 'gltf', '__init__': aframe_init})
gltf_model = type('gltf_model', (aframe_tag, Node, ParentNode), {'name': 'gltf_model', '__init__': aframe_init})
torus_knot = type('torus_knot', (aframe_tag, Node, ParentNode), {'name': 'torus_knot', '__init__': aframe_init})
obj_model = type('obj_model', (aframe_tag, Node, ParentNode), {'name': 'obj_model', '__init__': aframe_init})

# <a-box>
# <a-camera>
# <a-circle>
# <a-cone>
# <a-cursor>
# <a-curvedimage>
# <a-cylinder>
# <a-dodecahedron>
# <a-gltf-model>
# <a-icosahedron>
# <a-image>
# <a-light>
# <a-link>
# <a-obj-model>
# <a-octahedron>
# <a-plane>
# <a-ring>
# <a-sky>
# <a-sound>
# <a-sphere>
# <a-tetrahedron>
# <a-text>
# <a-torus-knot>
# <a-torus>
# <a-triangle>
# <a-video>
# <a-videosphere>



"""
    domonic.constants.color
    ====================================
"""

# from typing import Union, Tuple, List, Dict, Any, Optional, Callable, cast, TypeVar, Generic, Iterable, Sequence

from typing import Tuple

from domonic.geom import vec3, vec4

# function rgb
# function rgba
# function hsl
# function hsla
# function hex
# function hexa
# function hsv


class Color():
    """ Color Functions """

    # BLACK = Color(0, 0, 0)
    # WHITE = Color(255, 255, 255)

    @staticmethod
    def random_hex() -> str:
        """[returns a random hex color i.e. #000000]

        Returns:
            [str]: [random hex color i.e. #000000]
        """
        import random

        def r():
            return random.randint(0, 255)
        # r = lambda: random.randint(0, 255)
        return '#%02X%02X%02X' % (r(), r(), r())
        # import secrets
        # rgba = '#'+secrets.token_hex(4)

    @staticmethod
    def hex2rgb(h: str) -> Tuple[int, int, int]:
        """[takes a hex color in the form of #RRGGBB and returns the rgb values as a tuple i.e (r, g, b)]

        Args:
            h ([str]): [hex string i.e #ffffff]

        Returns:
            [tuple]: [rgb tuple i.e. (255, 255, 255)]
        """
        if h[0] == '#':
            h = h.lstrip('#')
        return tuple(int(h[i:i + 2], 16) for i in (0, 2, 4))

    # rgb = hex2rgb  # TODO - as static. but can also check if instance has value? so can get/set.

    # @staticmethod
    # def hsl2rgb(h, s, l):
    #     """ convert hsl to rgb """
    #     if s == 0:
    #         return l, l, l
    #     if l < 0.5:
    #         if s == 1:
    #             return l * (1 + l), l * 2, l * 2
    #         if s == 2:
    #             return l * 2, l * (1 + l), l * 2
    #         if s == 3:
    #             return l * 2, l * 2, l * (1 + l)
    #     else:
    #         if s == 1:
    #             return l * (1 - l), l * 2, l * 2
    #         if s == 2:
    #             return l * (1 - l), l * 2, l * (1 + l)
    #         if s == 3:
    #             return l * (1 - l), l * (1 + l), l * 2

    # @staticmethod
    # def rgb2hsl(r, g, b):
    #     """ convert rgb to hsl """
    #     maxv = max(r, g, b)
    #     minv = min(r, g, b)
    #     l = (maxv + minv) / 2
    #     if maxv == minv:
    #         return 0.0, 0.0, l
    #     s = (maxv - minv) / (maxv + minv)
    #     if r == maxv:
    #         h = (g - b) / (maxv - minv)
    #     elif g == maxv:
    #         h = 2 + (b - r) / (maxv - minv)
    #     else:
    #         h = 4 + (r - g) / (maxv - minv)
    #     h *= 60
    #     if h < 0:
    #         h += 360
    #     return h, s, l

    @staticmethod
    def rgb2hex(r: int, g: int, b: int) -> str:
        """[ takes 3 rgb values and returns a hex string i.e. #000000]

        Args:
            r ([int]): [a value between 0 and 255]
            g ([int]): [a value between 0 and 255]
            b ([int]): [a value between 0 and 255]

        Returns:
            [str]: [retuns a hex string i.e #ffffff]
        """
        #  TODO - pass tuples or
        # if isinstance(a, (int, float)):
        # elif isinstance(a, (tuple, list)):
        return '#%02x%02x%02x' % (r, g, b)

    # deprecated
    @staticmethod
    def fromRGBA(r: int, g: int, b: int, a: int = 255) -> 'Color':
        """[creates a Color from rgba values]

        Args:
            r ([int]): [a value between 0 and 255]
            g ([int]): [a value between 0 and 255]
            b ([int]): [a value between 0 and 255]
            a ([int]): [a value between 0 and 255]

        Returns:
            [type]: [a Color object]
        """
        return Color(r, g, b, a)

    # @staticmethod
    # def fromHsl(h, s, l):
    #     return Color(h, s, l)

    # deprecated
    @staticmethod
    def fromHex(hex: str) -> 'Color':
        """ create a Color from a hex string i.e. #ffffff """
        return Color(hex)

    def __init__(self, *args, **kwargs) -> None:
        """
        accepts a color in a variety of formats:
        """
        if isinstance(args[0], vec4):
            self.r = args[0][0]
            self.g = args[0][1]
            self.b = args[0][2]
            self.a = args[0][3]
        if isinstance(args[0], vec3):
            self.r = args[0][0]
            self.g = args[0][1]
            self.b = args[0][2]
        if isinstance(args[0], str):
            if args[0].startswith('#'):
                self.r, self.g, self.b = Color.hex2rgb(args[0])
        if isinstance(args[0], (int, float)):
            if len(args) == 3:
                self.r, self.g, self.b = args
            if len(args) == 4:
                self.r, self.g, self.b, self.a = args

        # self.alpha = kwargs.get('a', 1.0)
        # self.red = kwargs.get('r', self.r)
        # self.green = kwargs.get('g', self.g)
        # self.blue = kwargs.get('b', self.b)

        self.alpha = kwargs.get('alpha', 1.0)  # TODO - props instead?
        self.red = kwargs.get('red', self.r)
        self.green = kwargs.get('green', self.g)
        self.blue = kwargs.get('blue', self.b)
        # self.gray = kwargs.get('gray', self.r)
        self.hue = kwargs.get('hue', 0.0)
        self.saturation = kwargs.get('saturation', 1.0)
        self.brightness = kwargs.get('brightness', 1.0)
        self.lightness = kwargs.get('lightness', 1.0)

        # print(self.r, self.g, self.b)

    @property
    def alpha(self) -> float:
        return self.a

    @alpha.setter
    def alpha(self, value: float) -> None:
        self.a = value

    @property
    def red(self) -> float:
        return self.r

    @red.setter
    def red(self, value: float) -> None:
        self.r = value

    @property
    def green(self) -> float:
        return self.g

    @green.setter
    def green(self, value: float) -> None:
        self.g = value

    @property
    def blue(self) -> float:
        return self.b

    @blue.setter
    def blue(self, value: float) -> None:
        self.b = value

    def toRGB(self):  # : -> vec3:
        """[returns the color as RGB]

        Returns:
            [tuple]: [ (r, g, b) ]
        """
        return (self.r, self.g, self.b)

    def toHsl(self):
        """ returns the hsl for the color """
        return (self.hue, self.saturation, self.brightness)

    # def toString(self):
        # return str(self)

    def __str__(self) -> str:
        return Color.rgb2hex(self.r, self.g, self.b)

    # def __repr__(self):
    #     return str(self)

    def toHsv(self):
        """ get the hsv for the color """
        return (self.hue, self.saturation, self.brightness)

    def toCSS(self) -> str:
        """ return the color as a CSS string """
        return '#%02x%02x%02x' % (self.r, self.g, self.b)

    def toHex(self) -> str:
        return str(self)

    def toRGBA(self):
        return (self.r, self.g, self.b, self.a)

    def toSVG(self, shape="circle", size=10):
        """ returns the color as an svg string
        Args:
            shape ([str]): [can be circle or square]
            size ([int]): [size in pixels]
        """
        if shape == "circle":
            return '<circle cx="0" cy="0" r="%d" fill="%s" />' % (size, self.toHex())
        if shape == "square":
            return '<rect x="0" y="0" width="%d" height="%d" fill="%s" />' % (size, size, self.toHex())

    # def toIMG(self, size=10):
    #     """ returns the color as an svg string
    #     Args:
    #         shape ([str]): [can be circle or square]
    #         size ([int]): [size in pixels]
    #     """
    #     import PIL.Image as Image
    #     img = Image.new('RGB', (size, size), self.toHex())
    #     return img

    def convert(self, to: str):
        """ convert the color to a different color space

        Args:
            to ([str]): [can be one of the following: 'rgb', 'hsl', 'hsv', 'hex']
        """
        if to == 'rgb':
            return self.toRGB()
        if to == 'hsl':
            return self.toHsl()
        if to == 'hsv':
            return self.toHsv()
        if to == 'hex':
            return self.toHex()
        if to == 'css':
            return self.toCSS()

    # set(*args)

    def hasAlpha(self) -> bool:
        """[does the color have an alpha channel]

        Returns:
            [bool]: [True if alpha channel exists else False]
        """
        return self.a > 0

    def equals(self, color) -> bool:
        return self.r == color.r and self.g == color.g and self.b == color.b

    def __eq__(self, other):
        """ check if two colors are equal """
        return self.r == other.r and self.g == other.g and self.b == other.b

    def __add__(self, other):
        """ add two colors together """
        return Color(self.r + other.r, self.g + other.g, self.b + other.b)

    def __sub__(self, color):
        """ subtract a color from this color """
        return Color(self.r - color.r, self.g - color.g, self.b - color.b)

    def __mul__(self, color):
        """ multiply a color with this color """
        return Color(self.r * color.r, self.g * color.g, self.b * color.b)

    def __div__(self, color):
        """ divide a color with this color """
        return Color(self.r / color.r, self.g / color.g, self.b / color.b)

    # web
    Black: str = '#000000'  #:
    Navy: str = "#000080"  #:
    DarkBlue: str = "#00008B"  #:
    MediumBlue: str = "#0000CD"  #:
    Blue: str = "#0000FF"  #:
    DarkGreen: str = "#006400"  #:
    Green: str = "#008000"  #:
    Teal: str = "#008080"  #:
    DarkCyan: str = "#008B8B"  #:
    DeepSkyBlue: str = "#00BFFF"  #:
    DarkTurquoise: str = "#00CED1"  #:
    MediumSpringGreen: str = "#00FA9A"  #:
    Lime: str = "#00FF00"  #:
    SpringGreen: str = "#00FF7F"  #:
    Aqua: str = "#00FFFF"  #:
    Cyan: str = "#00FFFF"  #:
    MidnightBlue: str = "#191970"  #:
    DodgerBlue: str = "#1E90FF"  #:
    LightSeaGreen: str = "#20B2AA"  #:
    ForestGreen: str = "#228B22"  #:
    SeaGreen: str = "#2E8B57"  #:
    DarkSlateGray: str = "#2F4F4F"  #:
    DarkSlateGrey: str = "#2F4F4F"  #:
    LimeGreen: str = "#32CD32"  #:
    MediumSeaGreen: str = "#3CB371"  #:
    Turquoise: str = "#40E0D0"  #:
    RoyalBlue: str = "#4169E1"  #:
    SteelBlue: str = "#4682B4"  #:
    DarkSlateBlue: str = "#483D8B"  #:
    MediumTurquoise: str = "#48D1CC"  #:
    Indigo: str = "#4B0082"  #:
    DarkOliveGreen: str = "#556B2F"  #:
    CadetBlue: str = "#5F9EA0"  #:
    CornflowerBlue: str = "#6495ED"  #:
    RebeccaPurple: str = "#663399"  #:
    MediumAquaMarine: str = "#66CDAA"  #:
    DimGray: str = "#696969"  #:
    DimGrey: str = "#696969"  #:
    SlateBlue: str = "#6A5ACD"  #:
    OliveDrab: str = "#6B8E23"  #:
    SlateGray: str = "#708090"  #:
    SlateGrey: str = "#708090"  #:
    LightSlateGray: str = "#778899"  #:
    LightSlateGrey: str = "#778899"  #:
    MediumSlateBlue: str = "#7B68EE"  #:
    LawnGreen: str = "#7CFC00"  #:
    Chartreuse: str = "#7FFF00"  #:
    Aquamarine: str = "#7FFFD4"  #:
    Maroon: str = "#800000"  #:
    Purple: str = "#800080"  #:
    Olive: str = "#808000"  #:
    Gray: str = "#808080"  #:
    Grey: str = "#808080"  #:
    SkyBlue: str = "#87CEEB"  #:
    LightSkyBlue: str = "#87CEFA"  #:
    BlueViolet: str = "#8A2BE2"  #:
    DarkRed: str = "#8B0000"  #:
    DarkMagenta: str = "#8B008B"  #:
    SaddleBrown: str = "#8B4513"  #:
    DarkSeaGreen: str = "#8FBC8F"  #:
    LightGreen: str = "#90EE90"  #:
    MediumPurple: str = "#9370DB"  #:
    DarkViolet: str = "#9400D3"  #:
    PaleGreen: str = "#98FB98"  #:
    DarkOrchid: str = "#9932CC"  #:
    YellowGreen: str = "#9ACD32"  #:
    Sienna: str = "#A0522D"  #:
    Brown: str = "#A52A2A"  #:
    DarkGray: str = "#A9A9A9"  #:
    DarkGrey: str = "#A9A9A9"  #:
    LightBlue: str = "#ADD8E6"  #:
    GreenYellow: str = "#ADFF2F"  #:
    PaleTurquoise: str = "#AFEEEE"  #:
    LightSteelBlue: str = "#B0C4DE"  #:
    PowderBlue: str = "#B0E0E6"  #:
    FireBrick: str = "#B22222"  #:
    DarkGoldenRod: str = "#B8860B"  #:
    MediumOrchid: str = "#BA55D3"  #:
    RosyBrown: str = "#BC8F8F"  #:
    DarkKhaki: str = "#BDB76B"  #:
    Silver: str = "#C0C0C0"  #:
    MediumVioletRed: str = "#C71585"  #:
    IndianRed: str = "#CD5C5C"  #:
    Peru: str = "#CD853F"  #:
    Chocolate: str = "#D2691E"  #:
    Tan: str = "#D2B48C"  #:
    LightGray: str = "#D3D3D3"  #:
    LightGrey: str = "#D3D3D3"  #:
    Thistle: str = "#D8BFD8"  #:
    Orchid: str = "#DA70D6"  #:
    GoldenRod: str = "#DAA520"  #:
    PaleVioletRed: str = "#DB7093"  #:
    Crimson: str = "#DC143C"  #:
    Gainsboro: str = "#DCDCDC"  #:
    Plum: str = "#DDA0DD"  #:
    BurlyWood: str = "#DEB887"  #:
    LightCyan: str = "#E0FFFF"  #:
    Lavender: str = "#E6E6FA"  #:
    DarkSalmon: str = "#E9967A"  #:
    Violet: str = "#EE82EE"  #:
    PaleGoldenRod: str = "#EEE8AA"  #:
    LightCoral: str = "#F08080"  #:
    Khaki: str = "#F0E68C"  #:
    AliceBlue: str = "#F0F8FF"  #:
    HoneyDew: str = "#F0FFF0"  #:
    Azure: str = "#F0FFFF"  #:
    SandyBrown: str = "#F4A460"  #:
    Wheat: str = "#F5DEB3"  #:
    Beige: str = "#F5F5DC"  #:
    WhiteSmoke: str = "#F5F5F5"  #:
    MintCream: str = "#F5FFFA"  #:
    GhostWhite: str = "#F8F8FF"  #:
    Salmon: str = "#FA8072"  #:
    AntiqueWhite: str = "#FAEBD7"  #:
    Linen: str = "#FAF0E6"  #:
    LightGoldenRodYellow: str = "#FAFAD2"  #:
    OldLace: str = "#FDF5E6"  #:
    Red: str = "#FF0000"  #:
    Fuchsia: str = "#FF00FF"  #:
    Magenta: str = "#FF00FF"  #:
    DeepPink: str = "#FF1493"  #:
    OrangeRed: str = "#FF4500"  #:
    Tomato: str = "#FF6347"  #:
    HotPink: str = "#FF69B4"  #:
    Coral: str = "#FF7F50"  #:
    DarkOrange: str = "#FF8C00"  #:
    LightSalmon: str = "#FFA07A"  #:
    Orange: str = "#FFA500"  #:
    LightPink: str = "#FFB6C1"  #:
    Pink: str = "#FFC0CB"  #:
    Gold: str = "#FFD700"  #:
    PeachPuff: str = "#FFDAB9"  #:
    NavajoWhite: str = "#FFDEAD"  #:
    Moccasin: str = "#FFE4B5"  #:
    Bisque: str = "#FFE4C4"  #:
    MistyRose: str = "#FFE4E1"  #:
    BlanchedAlmond: str = "#FFEBCD"  #:
    PapayaWhip: str = "#FFEFD5"  #:
    LavenderBlush: str = "#FFF0F5"  #:
    SeaShell: str = "#FFF5EE"  #:
    Cornsilk: str = "#FFF8DC"  #:
    LemonChiffon: str = "#FFFACD"  #:
    FloralWhite: str = "#FFFAF0"  #:
    Snow: str = "#FFFAFA"  #:
    Yellow: str = "#FFFF00"  #:
    LightYellow: str = "#FFFFE0"  #:
    Ivory: str = "#FFFFF0"  #:
    White: str = "#FFFFFF"  #:

    # XKCD
    acidgreen: str = "#8ffe09"  #:
    adobe: str = "#bd6c48"  #:
    algae: str = "#54ac68"  #:
    algaegreen: str = "#21c36f"  #:
    almostblack: str = "#070d0d"  #:
    amber: str = "#feb308"  #:
    amethyst: str = "#9b5fc0"  #:
    apple: str = "#6ecb3c"  #:
    applegreen: str = "#76cd26"  #:
    apricot: str = "#ffb16d"  #:
    aqua: str = "#13eac9"  #:
    aquablue: str = "#02d8e9"  #:
    aquagreen: str = "#12e193"  #:
    aquamarine: str = "#2ee8bb"  #:
    armygreen: str = "#4b5d16"  #:
    asparagus: str = "#77ab56"  #:
    aubergine: str = "#3d0734"  #:
    auburn: str = "#9a3001"  #:
    avocado: str = "#90b134"  #:
    avocadogreen: str = "#87a922"  #:
    azul: str = "#1d5dec"  #:
    azure: str = "#069af3"  #:
    babyblue: str = "#a2cffe"  #:
    babygreen: str = "#8cff9e"  #:
    babypink: str = "#ffb7ce"  #:
    babypoo: str = "#ab9004"  #:
    babypoop: str = "#937c00"  #:
    babypoopgreen: str = "#8f9805"  #:
    babypukegreen: str = "#b6c406"  #:
    babypurple: str = "#ca9bf7"  #:
    babyshitbrown: str = "#ad900d"  #:
    babyshitgreen: str = "#889717"  #:
    banana: str = "#ffff7e"  #:
    bananayellow: str = "#fafe4b"  #:
    barbiepink: str = "#fe46a5"  #:
    barfgreen: str = "#94ac02"  #:
    barney: str = "#ac1db8"  #:
    barneypurple: str = "#a00498"  #:
    battleshipgrey: str = "#6b7c85"  #:
    beige: str = "#e6daa6"  #:
    berry: str = "#990f4b"  #:
    bile: str = "#b5c306"  #:
    black: str = "#000000"  #:
    bland: str = "#afa88b"  #:
    blood: str = "#770001"  #:
    bloodorange: str = "#fe4b03"  #:
    bloodred: str = "#980002"  #:
    blue: str = "#0343df"  #:
    blueberry: str = "#464196"  #:
    blueblue: str = "#2242c7"  #:
    bluegreen: str = "#0f9b8e"  #:
    bluegrey: str = "#85a3b2"  #:
    bluepurple: str = "#5a06ef"  #:
    blueviolet: str = "#5d06e9"  #:
    bluewithahintofpurple: str = "#533cc6"  #:
    blueygreen: str = "#2bb179"  #:
    blueygrey: str = "#89a0b0"  #:
    blueypurple: str = "#6241c7"  #:
    bluish: str = "#2976bb"  #:
    bluishgreen: str = "#10a674"  #:
    bluishgrey: str = "#748b97"  #:
    bluishpurple: str = "#703be7"  #:
    blurple: str = "#5539cc"  #:
    blush: str = "#f29e8e"  #:
    blushpink: str = "#fe828c"  #:
    booger: str = "#9bb53c"  #:
    boogergreen: str = "#96b403"  #:
    bordeaux: str = "#7b002c"  #:
    boringgreen: str = "#63b365"  #:
    bottlegreen: str = "#044a05"  #:
    brick: str = "#a03623"  #:
    brickorange: str = "#c14a09"  #:
    brickred: str = "#8f1402"  #:
    brightaqua: str = "#0bf9ea"  #:
    brightblue: str = "#0165fc"  #:
    brightcyan: str = "#41fdfe"  #:
    brightgreen: str = "#01ff07"  #:
    brightlavender: str = "#c760ff"  #:
    brightlightblue: str = "#26f7fd"  #:
    brightlightgreen: str = "#2dfe54"  #:
    brightlilac: str = "#c95efb"  #:
    brightlime: str = "#87fd05"  #:
    brightlimegreen: str = "#65fe08"  #:
    brightmagenta: str = "#ff08e8"  #:
    brightolive: str = "#9cbb04"  #:
    brightorange: str = "#ff5b00"  #:
    brightpink: str = "#fe01b1"  #:
    brightpurple: str = "#be03fd"  #:
    brightred: str = "#ff000d"  #:
    brightseagreen: str = "#05ffa6"  #:
    brightskyblue: str = "#02ccfe"  #:
    brightteal: str = "#01f9c6"  #:
    brightturquoise: str = "#0ffef9"  #:
    brightviolet: str = "#ad0afd"  #:
    brightyellow: str = "#fffd01"  #:
    brightyellowgreen: str = "#9dff00"  #:
    britishracinggreen: str = "#05480d"  #:
    bronze: str = "#a87900"  #:
    brown: str = "#653700"  #:
    browngreen: str = "#706c11"  #:
    browngrey: str = "#8d8468"  #:
    brownish: str = "#9c6d57"  #:
    brownishgreen: str = "#6a6e09"  #:
    brownishgrey: str = "#86775f"  #:
    brownishorange: str = "#cb7723"  #:
    brownishpink: str = "#c27e79"  #:
    brownishpurple: str = "#76424e"  #:
    brownishred: str = "#9e3623"  #:
    brownishyellow: str = "#c9b003"  #:
    brownorange: str = "#b96902"  #:
    brownred: str = "#922b05"  #:
    brownyellow: str = "#b29705"  #:
    brownygreen: str = "#6f6c0a"  #:
    brownyorange: str = "#ca6b02"  #:
    bruise: str = "#7e4071"  #:
    bubblegum: str = "#ff6cb5"  #:
    bubblegumpink: str = "#ff69af"  #:
    buff: str = "#fef69e"  #:
    burgundy: str = "#610023"  #:
    burntorange: str = "#c04e01"  #:
    burntred: str = "#9f2305"  #:
    burntsiena: str = "#b75203"  #:
    burntsienna: str = "#b04e0f"  #:
    burntumber: str = "#a0450e"  #:
    burntyellow: str = "#d5ab09"  #:
    burple: str = "#6832e3"  #:
    butter: str = "#ffff81"  #:
    butterscotch: str = "#fdb147"  #:
    butteryellow: str = "#fffd74"  #:
    cadetblue: str = "#4e7496"  #:
    camel: str = "#c69f59"  #:
    camo: str = "#7f8f4e"  #:
    camogreen: str = "#526525"  #:
    camouflagegreen: str = "#4b6113"  #:
    canary: str = "#fdff63"  #:
    canaryyellow: str = "#fffe40"  #:
    candypink: str = "#ff63e9"  #:
    caramel: str = "#af6f09"  #:
    carmine: str = "#9d0216"  #:
    carnation: str = "#fd798f"  #:
    carnationpink: str = "#ff7fa7"  #:
    carolinablue: str = "#8ab8fe"  #:
    celadon: str = "#befdb7"  #:
    celery: str = "#c1fd95"  #:
    cement: str = "#a5a391"  #:
    cerise: str = "#de0c62"  #:
    cerulean: str = "#0485d1"  #:
    ceruleanblue: str = "#056eee"  #:
    charcoal: str = "#343837"  #:
    charcoalgrey: str = "#3c4142"  #:
    chartreuse: str = "#c1f80a"  #:
    cherry: str = "#cf0234"  #:
    cherryred: str = "#f7022a"  #:
    chestnut: str = "#742802"  #:
    chocolate: str = "#3d1c02"  #:
    chocolatebrown: str = "#411900"  #:
    cinnamon: str = "#ac4f06"  #:
    claret: str = "#680018"  #:
    clay: str = "#b66a50"  #:
    claybrown: str = "#b2713d"  #:
    clearblue: str = "#247afd"  #:
    cobalt: str = "#1e488f"  #:
    cobaltblue: str = "#030aa7"  #:
    cocoa: str = "#875f42"  #:
    coffee: str = "#a6814c"  #:
    coolblue: str = "#4984b8"  #:
    coolgreen: str = "#33b864"  #:
    coolgrey: str = "#95a3a6"  #:
    copper: str = "#b66325"  #:
    coral: str = "#fc5a50"  #:
    coralpink: str = "#ff6163"  #:
    cornflower: str = "#6a79f7"  #:
    cornflowerblue: str = "#5170d7"  #:
    cranberry: str = "#9e003a"  #:
    cream: str = "#ffffc2"  #:
    creme: str = "#ffffb6"  #:
    crimson: str = "#8c000f"  #:
    custard: str = "#fffd78"  #:
    cyan: str = "#00ffff"  #:
    dandelion: str = "#fedf08"  #:
    dark: str = "#1b2431"  #:
    darkaqua: str = "#05696b"  #:
    darkaquamarine: str = "#017371"  #:
    darkbeige: str = "#ac9362"  #:
    darkblue: str = "#030764"  #:
    darkbluegreen: str = "#005249"  #:
    darkbluegrey: str = "#1f3b4d"  #:
    darkbrown: str = "#341c02"  #:
    darkcoral: str = "#cf524e"  #:
    darkcream: str = "#fff39a"  #:
    darkcyan: str = "#0a888a"  #:
    darkforestgreen: str = "#002d04"  #:
    darkfuchsia: str = "#9d0759"  #:
    darkgold: str = "#b59410"  #:
    darkgrassgreen: str = "#388004"  #:
    darkgreen: str = "#054907"  #:
    darkgreenblue: str = "#1f6357"  #:
    darkgrey: str = "#363737"  #:
    darkgreyblue: str = "#29465b"  #:
    darkhotpink: str = "#d90166"  #:
    darkindigo: str = "#1f0954"  #:
    darkishblue: str = "#014182"  #:
    darkishgreen: str = "#287c37"  #:
    darkishpink: str = "#da467d"  #:
    darkishpurple: str = "#751973"  #:
    darkishred: str = "#a90308"  #:
    darkkhaki: str = "#9b8f55"  #:
    darklavender: str = "#856798"  #:
    darklilac: str = "#9c6da5"  #:
    darklime: str = "#84b701"  #:
    darklimegreen: str = "#7ebd01"  #:
    darkmagenta: str = "#960056"  #:
    darkmaroon: str = "#3c0008"  #:
    darkmauve: str = "#874c62"  #:
    darkmint: str = "#48c072"  #:
    darkmintgreen: str = "#20c073"  #:
    darkmustard: str = "#a88905"  #:
    darknavy: str = "#000435"  #:
    darknavyblue: str = "#00022e"  #:
    darkolive: str = "#373e02"  #:
    darkolivegreen: str = "#3c4d03"  #:
    darkorange: str = "#c65102"  #:
    darkpastelgreen: str = "#56ae57"  #:
    darkpeach: str = "#de7e5d"  #:
    darkperiwinkle: str = "#665fd1"  #:
    darkpink: str = "#cb416b"  #:
    darkplum: str = "#3f012c"  #:
    darkpurple: str = "#35063e"  #:
    darkred: str = "#840000"  #:
    darkrose: str = "#b5485d"  #:
    darkroyalblue: str = "#02066f"  #:
    darksage: str = "#598556"  #:
    darksalmon: str = "#c85a53"  #:
    darksand: str = "#a88f59"  #:
    darkseafoam: str = "#1fb57a"  #:
    darkseafoamgreen: str = "#3eaf76"  #:
    darkseagreen: str = "#11875d"  #:
    darkskyblue: str = "#448ee4"  #:
    darkslateblue: str = "#214761"  #:
    darktan: str = "#af884a"  #:
    darktaupe: str = "#7f684e"  #:
    darkteal: str = "#014d4e"  #:
    darkturquoise: str = "#045c5a"  #:
    darkviolet: str = "#34013f"  #:
    darkyellow: str = "#d5b60a"  #:
    darkyellowgreen: str = "#728f02"  #:
    deepaqua: str = "#08787f"  #:
    deepblue: str = "#040273"  #:
    deepbrown: str = "#410200"  #:
    deepgreen: str = "#02590f"  #:
    deeplavender: str = "#8d5eb7"  #:
    deeplilac: str = "#966ebd"  #:
    deepmagenta: str = "#a0025c"  #:
    deeporange: str = "#dc4d01"  #:
    deeppink: str = "#cb0162"  #:
    deeppurple: str = "#36013f"  #:
    deepred: str = "#9a0200"  #:
    deeprose: str = "#c74767"  #:
    deepseablue: str = "#015482"  #:
    deepskyblue: str = "#0d75f8"  #:
    deepteal: str = "#00555a"  #:
    deepturquoise: str = "#017374"  #:
    deepviolet: str = "#490648"  #:
    denim: str = "#3b638c"  #:
    denimblue: str = "#3b5b92"  #:
    desert: str = "#ccad60"  #:
    diarrhea: str = "#9f8303"  #:
    dirt: str = "#8a6e45"  #:
    dirtbrown: str = "#836539"  #:
    dirtyblue: str = "#3f829d"  #:
    dirtygreen: str = "#667e2c"  #:
    dirtyorange: str = "#c87606"  #:
    dirtypink: str = "#ca7b80"  #:
    dirtypurple: str = "#734a65"  #:
    dirtyyellow: str = "#cdc50a"  #:
    dodgerblue: str = "#3e82fc"  #:
    drab: str = "#828344"  #:
    drabgreen: str = "#749551"  #:
    driedblood: str = "#4b0101"  #:
    duckeggblue: str = "#c3fbf4"  #:
    dullblue: str = "#49759c"  #:
    dullbrown: str = "#876e4b"  #:
    dullgreen: str = "#74a662"  #:
    dullorange: str = "#d8863b"  #:
    dullpink: str = "#d5869d"  #:
    dullpurple: str = "#84597e"  #:
    dullred: str = "#bb3f3f"  #:
    dullteal: str = "#5f9e8f"  #:
    dullyellow: str = "#eedc5b"  #:
    dusk: str = "#4e5481"  #:
    duskblue: str = "#26538d"  #:
    duskyblue: str = "#475f94"  #:
    duskypink: str = "#cc7a8b"  #:
    duskypurple: str = "#895b7b"  #:
    duskyrose: str = "#ba6873"  #:
    dust: str = "#b2996e"  #:
    dustyblue: str = "#5a86ad"  #:
    dustygreen: str = "#76a973"  #:
    dustylavender: str = "#ac86a8"  #:
    dustyorange: str = "#f0833a"  #:
    dustypink: str = "#d58a94"  #:
    dustypurple: str = "#825f87"  #:
    dustyred: str = "#b9484e"  #:
    dustyrose: str = "#c0737a"  #:
    dustyteal: str = "#4c9085"  #:
    earth: str = "#a2653e"  #:
    eastergreen: str = "#8cfd7e"  #:
    easterpurple: str = "#c071fe"  #:
    ecru: str = "#feffca"  #:
    eggplant: str = "#380835"  #:
    eggplantpurple: str = "#430541"  #:
    eggshell: str = "#fffcc4"  #:
    eggshellblue: str = "#c4fff7"  #:
    electricblue: str = "#0652ff"  #:
    electricgreen: str = "#21fc0d"  #:
    electriclime: str = "#a8ff04"  #:
    electricpink: str = "#ff0490"  #:
    electricpurple: str = "#aa23ff"  #:
    emerald: str = "#01a049"  #:
    emeraldgreen: str = "#028f1e"  #:
    evergreen: str = "#05472a"  #:
    fadedblue: str = "#658cbb"  #:
    fadedgreen: str = "#7bb274"  #:
    fadedorange: str = "#f0944d"  #:
    fadedpink: str = "#de9dac"  #:
    fadedpurple: str = "#916e99"  #:
    fadedred: str = "#d3494e"  #:
    fadedyellow: str = "#feff7f"  #:
    fawn: str = "#cfaf7b"  #:
    fern: str = "#63a950"  #:
    ferngreen: str = "#548d44"  #:
    fireenginered: str = "#fe0002"  #:
    flatblue: str = "#3c73a8"  #:
    flatgreen: str = "#699d4c"  #:
    fluorescentgreen: str = "#08ff08"  #:
    flurogreen: str = "#0aff02"  #:
    foamgreen: str = "#90fda9"  #:
    forest: str = "#0b5509"  #:
    forestgreen: str = "#06470c"  #:
    forrestgreen: str = "#154406"  #:
    frenchblue: str = "#436bad"  #:
    freshgreen: str = "#69d84f"  #:
    froggreen: str = "#58bc08"  #:
    fuchsia: str = "#ed0dd9"  #:
    gold: str = "#dbb40c"  #:
    golden: str = "#f5bf03"  #:
    goldenbrown: str = "#b27a01"  #:
    goldenrod: str = "#f9bc08"  #:
    goldenyellow: str = "#fec615"  #:
    grape: str = "#6c3461"  #:
    grapefruit: str = "#fd5956"  #:
    grapepurple: str = "#5d1451"  #:
    grass: str = "#5cac2d"  #:
    grassgreen: str = "#3f9b0b"  #:
    grassygreen: str = "#419c03"  #:
    green: str = "#15b01a"  #:
    greenapple: str = "#5edc1f"  #:
    greenblue: str = "#01c08d"  #:
    greenbrown: str = "#544e03"  #:
    greengrey: str = "#77926f"  #:
    greenish: str = "#40a368"  #:
    greenishbeige: str = "#c9d179"  #:
    greenishblue: str = "#0b8b87"  #:
    greenishbrown: str = "#696112"  #:
    greenishcyan: str = "#2afeb7"  #:
    greenishgrey: str = "#96ae8d"  #:
    greenishtan: str = "#bccb7a"  #:
    greenishteal: str = "#32bf84"  #:
    greenishturquoise: str = "#00fbb0"  #:
    greenishyellow: str = "#cdfd02"  #:
    greenteal: str = "#0cb577"  #:
    greenyblue: str = "#42b395"  #:
    greenybrown: str = "#696006"  #:
    greenyellow: str = "#b5ce08"  #:
    greenygrey: str = "#7ea07a"  #:
    greenyyellow: str = "#c6f808"  #:
    grey: str = "#929591"  #:
    greyblue: str = "#647d8e"  #:
    greybrown: str = "#7f7053"  #:
    greygreen: str = "#86a17d"  #:
    greyish: str = "#a8a495"  #:
    greyishblue: str = "#5e819d"  #:
    greyishbrown: str = "#7a6a4f"  #:
    greyishgreen: str = "#82a67d"  #:
    greyishpink: str = "#c88d94"  #:
    greyishpurple: str = "#887191"  #:
    greyishteal: str = "#719f91"  #:
    greypink: str = "#c3909b"  #:
    greypurple: str = "#826d8c"  #:
    greyteal: str = "#5e9b8a"  #:
    grossgreen: str = "#a0bf16"  #:
    gunmetal: str = "#536267"  #:
    hazel: str = "#8e7618"  #:
    heather: str = "#a484ac"  #:
    heliotrope: str = "#d94ff5"  #:
    highlightergreen: str = "#1bfc06"  #:
    hospitalgreen: str = "#9be5aa"  #:
    hotgreen: str = "#25ff29"  #:
    hotmagenta: str = "#f504c9"  #:
    hotpink: str = "#ff028d"  #:
    hotpurple: str = "#cb00f5"  #:
    huntergreen: str = "#0b4008"  #:
    ice: str = "#d6fffa"  #:
    iceblue: str = "#d7fffe"  #:
    ickygreen: str = "#8fae22"  #:
    indianred: str = "#850e04"  #:
    indigo: str = "#380282"  #:
    indigoblue: str = "#3a18b1"  #:
    iris: str = "#6258c4"  #:
    irishgreen: str = "#019529"  #:
    ivory: str = "#ffffcb"  #:
    jade: str = "#1fa774"  #:
    jadegreen: str = "#2baf6a"  #:
    junglegreen: str = "#048243"  #:
    kelleygreen: str = "#009337"  #:
    kellygreen: str = "#02ab2e"  #:
    kermitgreen: str = "#5cb200"  #:
    keylime: str = "#aeff6e"  #:
    khaki: str = "#aaa662"  #:
    khakigreen: str = "#728639"  #:
    kiwi: str = "#9cef43"  #:
    kiwigreen: str = "#8ee53f"  #:
    lavender: str = "#c79fef"  #:
    lavenderblue: str = "#8b88f8"  #:
    lavenderpink: str = "#dd85d7"  #:
    lawngreen: str = "#4da409"  #:
    leaf: str = "#71aa34"  #:
    leafgreen: str = "#5ca904"  #:
    leafygreen: str = "#51b73b"  #:
    leather: str = "#ac7434"  #:
    lemon: str = "#fdff52"  #:
    lemongreen: str = "#adf802"  #:
    lemonlime: str = "#bffe28"  #:
    lemonyellow: str = "#fdff38"  #:
    lichen: str = "#8fb67b"  #:
    lightaqua: str = "#8cffdb"  #:
    lightaquamarine: str = "#7bfdc7"  #:
    lightbeige: str = "#fffeb6"  #:
    lightblue: str = "#7bc8f6"  #:
    lightbluegreen: str = "#7efbb3"  #:
    lightbluegrey: str = "#b7c9e2"  #:
    lightbluishgreen: str = "#76fda8"  #:
    lightbrightgreen: str = "#53fe5c"  #:
    lightbrown: str = "#ad8150"  #:
    lightburgundy: str = "#a8415b"  #:
    lightcyan: str = "#acfffc"  #:
    lighteggplant: str = "#894585"  #:
    lightergreen: str = "#75fd63"  #:
    lighterpurple: str = "#a55af4"  #:
    lightforestgreen: str = "#4f9153"  #:
    lightgold: str = "#fddc5c"  #:
    lightgrassgreen: str = "#9af764"  #:
    lightgreen: str = "#76ff7b"  #:
    lightgreenblue: str = "#56fca2"  #:
    lightgreenishblue: str = "#63f7b4"  #:
    lightgrey: str = "#d8dcd6"  #:
    lightgreyblue: str = "#9dbcd4"  #:
    lightgreygreen: str = "#b7e1a1"  #:
    lightindigo: str = "#6d5acf"  #:
    lightishblue: str = "#3d7afd"  #:
    lightishgreen: str = "#61e160"  #:
    lightishpurple: str = "#a552e6"  #:
    lightishred: str = "#fe2f4a"  #:
    lightkhaki: str = "#e6f2a2"  #:
    lightlavendar: str = "#efc0fe"  #:
    lightlavender: str = "#dfc5fe"  #:
    lightlightblue: str = "#cafffb"  #:
    lightlightgreen: str = "#c8ffb0"  #:
    lightlilac: str = "#edc8ff"  #:
    lightlime: str = "#aefd6c"  #:
    lightlimegreen: str = "#b9ff66"  #:
    lightmagenta: str = "#fa5ff7"  #:
    lightmaroon: str = "#a24857"  #:
    lightmauve: str = "#c292a1"  #:
    lightmint: str = "#b6ffbb"  #:
    lightmintgreen: str = "#a6fbb2"  #:
    lightmossgreen: str = "#a6c875"  #:
    lightmustard: str = "#f7d560"  #:
    lightnavy: str = "#155084"  #:
    lightnavyblue: str = "#2e5a88"  #:
    lightneongreen: str = "#4efd54"  #:
    lightolive: str = "#acbf69"  #:
    lightolivegreen: str = "#a4be5c"  #:
    lightorange: str = "#fdaa48"  #:
    lightpastelgreen: str = "#b2fba5"  #:
    lightpeach: str = "#ffd8b1"  #:
    lightpeagreen: str = "#c4fe82"  #:
    lightperiwinkle: str = "#c1c6fc"  #:
    lightpink: str = "#ffd1df"  #:
    lightplum: str = "#9d5783"  #:
    lightpurple: str = "#bf77f6"  #:
    lightred: str = "#ff474c"  #:
    lightrose: str = "#ffc5cb"  #:
    lightroyalblue: str = "#3a2efe"  #:
    lightsage: str = "#bcecac"  #:
    lightsalmon: str = "#fea993"  #:
    lightseafoam: str = "#a0febf"  #:
    lightseafoamgreen: str = "#a7ffb5"  #:
    lightseagreen: str = "#98f6b0"  #:
    lightskyblue: str = "#c6fcff"  #:
    lighttan: str = "#fbeeac"  #:
    lightteal: str = "#90e4c1"  #:
    lightturquoise: str = "#7ef4cc"  #:
    lighturple: str = "#b36ff6"  #:
    lightviolet: str = "#d6b4fc"  #:
    lightyellow: str = "#fffe7a"  #:
    lightyellowgreen: str = "#ccfd7f"  #:
    lightyellowishgreen: str = "#c2ff89"  #:
    lilac: str = "#cea2fd"  #:
    liliac: str = "#c48efd"  #:
    lime: str = "#aaff32"  #:
    limegreen: str = "#89fe05"  #:
    limeyellow: str = "#d0fe1d"  #:
    lipstick: str = "#d5174e"  #:
    lipstickred: str = "#c0022f"  #:
    macaroniandcheese: str = "#efb435"  #:
    magenta: str = "#c20078"  #:
    mahogany: str = "#4a0100"  #:
    maize: str = "#f4d054"  #:
    mango: str = "#ffa62b"  #:
    manilla: str = "#fffa86"  #:
    marigold: str = "#fcc006"  #:
    marine: str = "#042e60"  #:
    marineblue: str = "#01386a"  #:
    maroon: str = "#650021"  #:
    mauve: str = "#ae7181"  #:
    mediumblue: str = "#2c6fbb"  #:
    mediumbrown: str = "#7f5112"  #:
    mediumgreen: str = "#39ad48"  #:
    mediumgrey: str = "#7d7f7c"  #:
    mediumpink: str = "#f36196"  #:
    mediumpurple: str = "#9e43a2"  #:
    melon: str = "#ff7855"  #:
    merlot: str = "#730039"  #:
    metallicblue: str = "#4f738e"  #:
    midblue: str = "#276ab3"  #:
    midgreen: str = "#50a747"  #:
    midnight: str = "#03012d"  #:
    midnightblue: str = "#020035"  #:
    midnightpurple: str = "#280137"  #:
    militarygreen: str = "#667c3e"  #:
    milkchocolate: str = "#7f4e1e"  #:
    mint: str = "#9ffeb0"  #:
    mintgreen: str = "#8fff9f"  #:
    mintygreen: str = "#0bf77d"  #:
    mocha: str = "#9d7651"  #:
    moss: str = "#769958"  #:
    mossgreen: str = "#658b38"  #:
    mossygreen: str = "#638b27"  #:
    mud: str = "#735c12"  #:
    mudbrown: str = "#60460f"  #:
    muddybrown: str = "#886806"  #:
    muddygreen: str = "#657432"  #:
    muddyyellow: str = "#bfac05"  #:
    mudgreen: str = "#606602"  #:
    mulberry: str = "#920a4e"  #:
    murkygreen: str = "#6c7a0e"  #:
    mushroom: str = "#ba9e88"  #:
    mustard: str = "#ceb301"  #:
    mustardbrown: str = "#ac7e04"  #:
    mustardgreen: str = "#a8b504"  #:
    mustardyellow: str = "#d2bd0a"  #:
    mutedblue: str = "#3b719f"  #:
    mutedgreen: str = "#5fa052"  #:
    mutedpink: str = "#d1768f"  #:
    mutedpurple: str = "#805b87"  #:
    nastygreen: str = "#70b23f"  #:
    navy: str = "#01153e"  #:
    navyblue: str = "#001146"  #:
    navygreen: str = "#35530a"  #:
    neonblue: str = "#04d9ff"  #:
    neongreen: str = "#0cff0c"  #:
    neonpink: str = "#fe019a"  #:
    neonpurple: str = "#bc13fe"  #:
    neonred: str = "#ff073a"  #:
    neonyellow: str = "#cfff04"  #:
    niceblue: str = "#107ab0"  #:
    nightblue: str = "#040348"  #:
    ocean: str = "#017b92"  #:
    oceanblue: str = "#03719c"  #:
    oceangreen: str = "#3d9973"  #:
    ocher: str = "#bf9b0c"  #:
    ochre: str = "#bf9005"  #:
    ocre: str = "#c69c04"  #:
    offblue: str = "#5684ae"  #:
    offgreen: str = "#6ba353"  #:
    offwhite: str = "#ffffe4"  #:
    offyellow: str = "#f1f33f"  #:
    oldpink: str = "#c77986"  #:
    oldrose: str = "#c87f89"  #:
    olive: str = "#6e750e"  #:
    olivebrown: str = "#645403"  #:
    olivedrab: str = "#6f7632"  #:
    olivegreen: str = "#677a04"  #:
    oliveyellow: str = "#c2b709"  #:
    orange: str = "#f97306"  #:
    orangebrown: str = "#be6400"  #:
    orangeish: str = "#fd8d49"  #:
    orangepink: str = "#ff6f52"  #:
    orangered: str = "#fe420f"  #:
    orangeybrown: str = "#b16002"  #:
    orangeyellow: str = "#ffad01"  #:
    orangeyred: str = "#fa4224"  #:
    orangeyyellow: str = "#fdb915"  #:
    orangish: str = "#fc824a"  #:
    orangishbrown: str = "#b25f03"  #:
    orangishred: str = "#f43605"  #:
    orchid: str = "#c875c4"  #:
    pale: str = "#fff9d0"  #:
    paleaqua: str = "#b8ffeb"  #:
    paleblue: str = "#d0fefe"  #:
    palebrown: str = "#b1916e"  #:
    palecyan: str = "#b7fffa"  #:
    palegold: str = "#fdde6c"  #:
    palegreen: str = "#c7fdb5"  #:
    palegrey: str = "#fdfdfe"  #:
    palelavender: str = "#eecffe"  #:
    palelightgreen: str = "#b1fc99"  #:
    palelilac: str = "#e4cbff"  #:
    palelime: str = "#befd73"  #:
    palelimegreen: str = "#b1ff65"  #:
    palemagenta: str = "#d767ad"  #:
    palemauve: str = "#fed0fc"  #:
    paleolive: str = "#b9cc81"  #:
    paleolivegreen: str = "#b1d27b"  #:
    paleorange: str = "#ffa756"  #:
    palepeach: str = "#ffe5ad"  #:
    palepink: str = "#ffcfdc"  #:
    palepurple: str = "#b790d4"  #:
    palered: str = "#d9544d"  #:
    palerose: str = "#fdc1c5"  #:
    palesalmon: str = "#ffb19a"  #:
    paleskyblue: str = "#bdf6fe"  #:
    paleteal: str = "#82cbb2"  #:
    paleturquoise: str = "#a5fbd5"  #:
    paleviolet: str = "#ceaefa"  #:
    paleyellow: str = "#ffff84"  #:
    parchment: str = "#fefcaf"  #:
    pastelblue: str = "#a2bffe"  #:
    pastelgreen: str = "#b0ff9d"  #:
    pastelorange: str = "#ff964f"  #:
    pastelpink: str = "#ffbacd"  #:
    pastelpurple: str = "#caa0ff"  #:
    pastelred: str = "#db5856"  #:
    pastelyellow: str = "#fffe71"  #:
    pea: str = "#a4bf20"  #:
    peach: str = "#ffb07c"  #:
    peachypink: str = "#ff9a8a"  #:
    peacockblue: str = "#016795"  #:
    peagreen: str = "#8eab12"  #:
    pear: str = "#cbf85f"  #:
    peasoup: str = "#929901"  #:
    peasoupgreen: str = "#94a617"  #:
    periwinkle: str = "#8e82fe"  #:
    periwinkleblue: str = "#8f99fb"  #:
    perrywinkle: str = "#8f8ce7"  #:
    petrol: str = "#005f6a"  #:
    pigpink: str = "#e78ea5"  #:
    pine: str = "#2b5d34"  #:
    pinegreen: str = "#0a481e"  #:
    pink: str = "#ff81c0"  #:
    pinkish: str = "#d46a7e"  #:
    pinkishbrown: str = "#b17261"  #:
    pinkishgrey: str = "#c8aca9"  #:
    pinkishorange: str = "#ff724c"  #:
    pinkishpurple: str = "#d648d7"  #:
    pinkishred: str = "#f10c45"  #:
    pinkishtan: str = "#d99b82"  #:
    pinkpurple: str = "#ef1de7"  #:
    pinkred: str = "#f5054f"  #:
    pinky: str = "#fc86aa"  #:
    pinkypurple: str = "#c94cbe"  #:
    pinkyred: str = "#fc2647"  #:
    pissyellow: str = "#ddd618"  #:
    pistachio: str = "#c0fa8b"  #:
    plum: str = "#580f41"  #:
    plumpurple: str = "#4e0550"  #:
    poisongreen: str = "#40fd14"  #:
    poo: str = "#8f7303"  #:
    poobrown: str = "#885f01"  #:
    poop: str = "#7f5e00"  #:
    poopbrown: str = "#7a5901"  #:
    poopgreen: str = "#6f7c00"  #:
    powderblue: str = "#b1d1fc"  #:
    powderpink: str = "#ffb2d0"  #:
    primaryblue: str = "#0804f9"  #:
    prussianblue: str = "#004577"  #:
    puce: str = "#a57e52"  #:
    puke: str = "#a5a502"  #:
    pukebrown: str = "#947706"  #:
    pukegreen: str = "#9aae07"  #:
    pukeyellow: str = "#c2be0e"  #:
    pumpkin: str = "#e17701"  #:
    pumpkinorange: str = "#fb7d07"  #:
    pureblue: str = "#0203e2"  #:
    purple: str = "#7e1e9c"  #:
    purpleblue: str = "#5d21d0"  #:
    purplebrown: str = "#673a3f"  #:
    purplegrey: str = "#866f85"  #:
    purpleish: str = "#98568d"  #:
    purpleishblue: str = "#6140ef"  #:
    purpleishpink: str = "#df4ec8"  #:
    purplepink: str = "#d725de"  #:
    purplered: str = "#990147"  #:
    purpley: str = "#8756e4"  #:
    purpleyblue: str = "#5f34e7"  #:
    purpleygrey: str = "#947e94"  #:
    purpleypink: str = "#c83cb9"  #:
    purplish: str = "#94568c"  #:
    purplishblue: str = "#601ef9"  #:
    purplishbrown: str = "#6b4247"  #:
    purplishgrey: str = "#7a687f"  #:
    purplishpink: str = "#ce5dae"  #:
    purplishred: str = "#b0054b"  #:
    purply: str = "#983fb2"  #:
    purplyblue: str = "#661aee"  #:
    purplypink: str = "#f075e6"  #:
    putty: str = "#beae8a"  #:
    racinggreen: str = "#014600"  #:
    radioactivegreen: str = "#2cfa1f"  #:
    raspberry: str = "#b00149"  #:
    rawsienna: str = "#9a6200"  #:
    rawumber: str = "#a75e09"  #:
    reallylightblue: str = "#d4ffff"  #:
    red: str = "#e50000"  #:
    redbrown: str = "#8b2e16"  #:
    reddish: str = "#c44240"  #:
    reddishbrown: str = "#7f2b0a"  #:
    reddishgrey: str = "#997570"  #:
    reddishorange: str = "#f8481c"  #:
    reddishpink: str = "#fe2c54"  #:
    reddishpurple: str = "#910951"  #:
    reddybrown: str = "#6e1005"  #:
    redorange: str = "#fd3c06"  #:
    redpink: str = "#fa2a55"  #:
    redpurple: str = "#820747"  #:
    redviolet: str = "#9e0168"  #:
    redwine: str = "#8c0034"  #:
    richblue: str = "#021bf9"  #:
    richpurple: str = "#720058"  #:
    robineggblue: str = "#8af1fe"  #:
    robinsegg: str = "#6dedfd"  #:
    robinseggblue: str = "#98eff9"  #:
    rosa: str = "#fe86a4"  #:
    rose: str = "#cf6275"  #:
    rosepink: str = "#f7879a"  #:
    rosered: str = "#be013c"  #:
    rosypink: str = "#f6688e"  #:
    rouge: str = "#ab1239"  #:
    royal: str = "#0c1793"  #:
    royalblue: str = "#0504aa"  #:
    royalpurple: str = "#4b006e"  #:
    ruby: str = "#ca0147"  #:
    russet: str = "#a13905"  #:
    rust: str = "#a83c09"  #:
    rustbrown: str = "#8b3103"  #:
    rustorange: str = "#c45508"  #:
    rustred: str = "#aa2704"  #:
    rustyorange: str = "#cd5909"  #:
    rustyred: str = "#af2f0d"  #:
    saffron: str = "#feb209"  #:
    sage: str = "#87ae73"  #:
    sagegreen: str = "#88b378"  #:
    salmon: str = "#ff796c"  #:
    salmonpink: str = "#fe7b7c"  #:
    sand: str = "#e2ca76"  #:
    sandbrown: str = "#cba560"  #:
    sandstone: str = "#c9ae74"  #:
    sandy: str = "#f1da7a"  #:
    sandybrown: str = "#c4a661"  #:
    sandyellow: str = "#fce166"  #:
    sandyyellow: str = "#fdee73"  #:
    sapgreen: str = "#5c8b15"  #:
    sapphire: str = "#2138ab"  #:
    scarlet: str = "#be0119"  #:
    sea: str = "#3c9992"  #:
    seablue: str = "#047495"  #:
    seafoam: str = "#80f9ad"  #:
    seafoamblue: str = "#78d1b6"  #:
    seafoamgreen: str = "#7af9ab"  #:
    seagreen: str = "#53fca1"  #:
    seaweed: str = "#18d17b"  #:
    seaweedgreen: str = "#35ad6b"  #:
    sepia: str = "#985e2b"  #:
    shamrock: str = "#01b44c"  #:
    shamrockgreen: str = "#02c14d"  #:
    shit: str = "#7f5f00"  #:
    shitbrown: str = "#7b5804"  #:
    shitgreen: str = "#758000"  #:
    shockingpink: str = "#fe02a2"  #:
    sickgreen: str = "#9db92c"  #:
    sicklygreen: str = "#94b21c"  #:
    sicklyyellow: str = "#d0e429"  #:
    sienna: str = "#a9561e"  #:
    silver: str = "#c5c9c7"  #:
    sky: str = "#82cafc"  #:
    skyblue: str = "#75bbfd"  #:
    slate: str = "#516572"  #:
    slateblue: str = "#5b7c99"  #:
    slategreen: str = "#658d6d"  #:
    slategrey: str = "#59656d"  #:
    slimegreen: str = "#99cc04"  #:
    snot: str = "#acbb0d"  #:
    snotgreen: str = "#9dc100"  #:
    softblue: str = "#6488ea"  #:
    softgreen: str = "#6fc276"  #:
    softpink: str = "#fdb0c0"  #:
    softpurple: str = "#a66fb5"  #:
    spearmint: str = "#1ef876"  #:
    springgreen: str = "#a9f971"  #:
    spruce: str = "#0a5f38"  #:
    squash: str = "#f2ab15"  #:
    steel: str = "#738595"  #:
    steelblue: str = "#5a7d9a"  #:
    steelgrey: str = "#6f828a"  #:
    stone: str = "#ada587"  #:
    stormyblue: str = "#507b9c"  #:
    straw: str = "#fcf679"  #:
    strawberry: str = "#fb2943"  #:
    strongblue: str = "#0c06f7"  #:
    strongpink: str = "#ff0789"  #:
    sunflower: str = "#ffc512"  #:
    sunfloweryellow: str = "#ffda03"  #:
    sunnyyellow: str = "#fff917"  #:
    sunshineyellow: str = "#fffd37"  #:
    sunyellow: str = "#ffdf22"  #:
    swamp: str = "#698339"  #:
    swampgreen: str = "#748500"  #:
    tan: str = "#d1b26f"  #:
    tanbrown: str = "#ab7e4c"  #:
    tangerine: str = "#ff9408"  #:
    tangreen: str = "#a9be70"  #:
    taupe: str = "#b9a281"  #:
    tea: str = "#65ab7c"  #:
    teagreen: str = "#bdf8a3"  #:
    teal: str = "#029386"  #:
    tealblue: str = "#01889f"  #:
    tealgreen: str = "#25a36f"  #:
    tealish: str = "#24bca8"  #:
    tealishgreen: str = "#0cdc73"  #:
    terracota: str = "#cb6843"  #:
    terracotta: str = "#c9643b"  #:
    tiffanyblue: str = "#7bf2da"  #:
    tomato: str = "#ef4026"  #:
    tomatored: str = "#ec2d01"  #:
    topaz: str = "#13bbaf"  #:
    toupe: str = "#c7ac7d"  #:
    toxicgreen: str = "#61de2a"  #:
    treegreen: str = "#2a7e19"  #:
    trueblue: str = "#010fcc"  #:
    truegreen: str = "#089404"  #:
    turquoise: str = "#06c2ac"  #:
    turquoiseblue: str = "#06b1c4"  #:
    turquoisegreen: str = "#04f489"  #:
    turtlegreen: str = "#75b84f"  #:
    twilight: str = "#4e518b"  #:
    twilightblue: str = "#0a437a"  #:
    uglyblue: str = "#31668a"  #:
    uglybrown: str = "#7d7103"  #:
    uglygreen: str = "#7a9703"  #:
    uglypink: str = "#cd7584"  #:
    uglypurple: str = "#a442a0"  #:
    uglyyellow: str = "#d0c101"  #:
    ultramarine: str = "#2000b1"  #:
    ultramarineblue: str = "#1805db"  #:
    umber: str = "#b26400"  #:
    velvet: str = "#750851"  #:
    vermillion: str = "#f4320c"  #:
    verydarkblue: str = "#000133"  #:
    verydarkbrown: str = "#1d0200"  #:
    verydarkgreen: str = "#062e03"  #:
    verydarkpurple: str = "#2a0134"  #:
    verylightblue: str = "#d5ffff"  #:
    verylightbrown: str = "#d3b683"  #:
    verylightgreen: str = "#d1ffbd"  #:
    verylightpink: str = "#fff4f2"  #:
    verylightpurple: str = "#f6cefc"  #:
    verypaleblue: str = "#d6fffe"  #:
    verypalegreen: str = "#cffdbc"  #:
    vibrantblue: str = "#0339f8"  #:
    vibrantgreen: str = "#0add08"  #:
    vibrantpurple: str = "#ad03de"  #:
    violet: str = "#9a0eea"  #:
    violetblue: str = "#510ac9"  #:
    violetpink: str = "#fb5ffc"  #:
    violetred: str = "#a50055"  #:
    viridian: str = "#1e9167"  #:
    vividblue: str = "#152eff"  #:
    vividgreen: str = "#2fef10"  #:
    vividpurple: str = "#9900fa"  #:
    vomit: str = "#a2a415"  #:
    vomitgreen: str = "#89a203"  #:
    vomityellow: str = "#c7c10c"  #:
    warmblue: str = "#4b57db"  #:
    warmbrown: str = "#964e02"  #:
    warmgrey: str = "#978a84"  #:
    warmpink: str = "#fb5581"  #:
    warmpurple: str = "#952e8f"  #:
    washedoutgreen: str = "#bcf5a6"  #:
    waterblue: str = "#0e87cc"  #:
    watermelon: str = "#fd4659"  #:
    weirdgreen: str = "#3ae57f"  #:
    wheat: str = "#fbdd7e"  #:
    white: str = "#ffffff"  #:
    windowsblue: str = "#3778bf"  #:
    wine: str = "#80013f"  #:
    winered: str = "#7b0323"  #:
    wintergreen: str = "#20f986"  #:
    wisteria: str = "#a87dc2"  #:
    yellow: str = "#ffff14"  #:
    yellowbrown: str = "#b79400"  #:
    yellowgreen: str = "#bbf90f"  #:
    yellowish: str = "#faee66"  #:
    yellowishbrown: str = "#9b7a01"  #:
    yellowishgreen: str = "#b0dd16"  #:
    yellowishorange: str = "#ffab0f"  #:
    yellowishtan: str = "#fcfc81"  #:
    yellowochre: str = "#cb9d06"  #:
    yelloworange: str = "#fcb001"  #:
    yellowtan: str = "#ffe36e"  #:
    yellowybrown: str = "#ae8b0c"  #:
    yellowygreen: str = "#bff128"  #:
